/*
 * arch/arm/mach-ns9xxx/gpio.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/ioport.h>

#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/io.h>

#include <asm/arch/regs-bbus.h>
#include <asm/arch/hardware.h>


#define GPIO_MAX_NR	72	


int ns9xxx_request_gpio(unsigned int pin, const char *name)
{
	if(pin > GPIO_MAX_NR) {
		printk( KERN_WARNING "Unavailable GPIO number %d (max GPIO is %d)\n", 
			pin, GPIO_MAX_NR );
			return -EINVAL;
	}
	if(request_region(REQUEST_GPIO_BASE_VA+pin, 1, name) == NULL)
		return -EBUSY;

	return 0;
}
EXPORT_SYMBOL(ns9xxx_request_gpio);


void ns9xxx_release_gpio(unsigned int pin)
{
	if(pin > GPIO_MAX_NR) {
		printk( KERN_WARNING "Unavailable GPIO number %d (max GPIO is %d)\n", 
			pin, GPIO_MAX_NR );
			return;
	}
	release_region((REQUEST_GPIO_BASE_VA+pin), 1);
}
EXPORT_SYMBOL(ns9xxx_release_gpio);


int ns9xxx_gpio_cfgpin(unsigned int pin, unsigned int function)
{
	void __iomem *addr = (void __iomem *)(NS9XXX_CFG_ADDR_BASE_VA + NS9XXX_CFG_PIN_OFFSET(pin));
	u32 bitpos = (((pin) % 8) * 4);
	u32 mask = 0x0f << bitpos;
	u32 cfgreg;
	unsigned long flags;

	if(pin > GPIO_MAX_NR) {
		printk( KERN_WARNING "Unavailable GPIO number %d (max GPIO is %d)\n", 
			pin, GPIO_MAX_NR );
			return -EINVAL;
	}

	switch (function & ~GPIO_CFG_INVERT ) {
		case GPIO_CFG_FUNC_0:
		case GPIO_CFG_FUNC_1:
		case GPIO_CFG_FUNC_2:
		case GPIO_CFG_INPUT:
		case GPIO_CFG_OUTPUT:
			/* just for input-check, default will return */
			break;
		
		default:
			printk( KERN_WARNING "Unknown function (0x%08x) for GPIO pin %d\n", 
				function, pin );
			return -EINVAL;
	}

	local_irq_save(flags);
	cfgreg = __raw_readl(addr) & ~mask;
	cfgreg |= (function << bitpos);
	__raw_writel(cfgreg, addr);
 	local_irq_restore(flags);

	return 0;
}
EXPORT_SYMBOL(ns9xxx_gpio_cfgpin);


u32 ns9xxx_gpio_getcfg(unsigned int pin)
{
	void __iomem *addr = (void __iomem *)(NS9XXX_CFG_ADDR_BASE_VA + NS9XXX_CFG_PIN_OFFSET(pin));
	u32 mask = 0x0f << (((pin) % 8) * 4);

	return __raw_readl(addr) & mask;
}
EXPORT_SYMBOL(ns9xxx_gpio_getcfg);


int ns9xxx_gpio_invert(unsigned int pin, unsigned int val)
{
	void __iomem *addr = (void __iomem *)(NS9XXX_CFG_ADDR_BASE_VA + NS9XXX_CFG_PIN_OFFSET(pin));
	unsigned long flags;
	u32 cfgreg;
	u32 mask = 0x04 << (((pin) % 8) * 4);

	if(pin > GPIO_MAX_NR) {
		printk( KERN_WARNING "Unavailable GPIO number %d (max GPIO is %d)\n", 
			pin, GPIO_MAX_NR );
			return -EINVAL;
	}

	local_irq_save(flags);
	cfgreg = __raw_readl(addr) & ~mask;
	if(val)
		cfgreg |= mask;
	__raw_writel(cfgreg, addr);
	local_irq_restore(flags);

	return 0;
}
EXPORT_SYMBOL(ns9xxx_gpio_invert);


int ns9xxx_gpio_setpin(unsigned int pin, unsigned int val)
{
	void __iomem *addr = (void __iomem *)(NS9XXX_CTRL_ADDR_BASE_VA + NS9XXX_CTRL_PIN_OFFSET(pin));
	unsigned long flags;
	u32 ctrlreg;
	u32 mask = (1<<((pin)%32));

	if(pin > GPIO_MAX_NR) {
		printk( KERN_WARNING "Unavailable GPIO number %d (max GPIO is %d)\n", 
			pin, GPIO_MAX_NR );
			return -EINVAL;
	}

	local_irq_save(flags);
	ctrlreg = __raw_readl(addr) & ~mask;
	if(val)
		ctrlreg |= mask;
	__raw_writel(ctrlreg, addr);
	local_irq_restore(flags);

	return 0;
}
EXPORT_SYMBOL(ns9xxx_gpio_setpin);


unsigned int ns9xxx_gpio_getpin(unsigned int pin)
{
	void __iomem *addr = (void __iomem *)(NS9XXX_STAT_ADDR_BASE_VA + NS9XXX_STAT_PIN_OFFSET(pin));

	/* XXX use and mask?? should be faster? */
	return __raw_readl(addr) & (1<<((pin)%32));
}
EXPORT_SYMBOL(ns9xxx_gpio_getpin);

