/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: ns9750-pci.c,v 1.1 2007-03-01 08:33:32 mpietrek Exp $
 * @Author: Leonard Samuelson
 * @Descr: PCI support for NetSilicon NS9750 SOC
 * @References: [1]
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Much of this code contains ideas derived from the NS9750 Ethernet
 * driver: .../drivers/net/arm/ns9750_eth.c
 *
 ***********************************************************************/

#include <linux/config.h>	/* custom configuration */

#include <linux/init.h>		/* __init */
#include <linux/pci.h>
#include <linux/interrupt.h>
#ifdef CONFIG_PCI_PROC_ERRORS
# include <linux/proc_fs.h>
#endif
#include <linux/ptrace.h>

#include <asm/io.h>
#include <asm/arch/irqs.h>
#include <asm/arch-ns9750/ns9750_sys.h>		/* NS_SYS_RESET */
#include <asm/arch-ns9750/ns9750_pci.h>
#ifdef CONFIG_PCI_PROC_ERRORS
# include <asm/page.h>				/* PAGE_SIZE */
#endif
#include <asm/pci.h>
#include <asm/mach/pci.h>

#undef DEBUG_NSPCI
// #define DEBUG_NSPCI  1

#ifdef	DEBUG_NSPCI
# define DEBUG_INIT		0x0001
# define DEBUG_MINOR		0x0002
# define DEBUG_CONFIG		0x0004
# define DEBUG_BUSINIT		0x0008
# define DEBUG_INFO		0x0010
# define DEBUG_VERBOSE		0x0020
# define DEBUG_ERROR		0x4000
# define DEBUG_ERROR_CRIT	0x8000

# define DEBUG_ALL		0xFFFF

static int debug_level = DEBUG_ALL;

# define DEBUG_ARGS0( FLG, a0 ) if( ( debug_level & (FLG) ) == (FLG) ) \
		printk( KERN_INFO "%s: " a0,  __FUNCTION__ )
# define DEBUG_ARGS1( FLG, a0, a1 ) if( ( debug_level & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, (a1) )
# define DEBUG_ARGS2( FLG, a0, a1, a2 ) if( (debug_level & (FLG)) ==(FLG))\
		printk( KERN_INFO "%s: " a0, __FUNCTION__, (a1), (a2) )
# define DEBUG_ARGS3( FLG, a0, a1, a2, a3 ) if((debug_level &(FLG))==(FLG))\
		printk(KERN_INFO "%s: " a0, __FUNCTION__,(a1),(a2),(a3))
# define DEBUG_ARGS4( FLG, a0, a1, a2, a3, a4 ) if((debug_level &(FLG))==(FLG))\
		printk(KERN_INFO "%s: " a0, __FUNCTION__,(a1),(a2),(a3),(a4))
# define DEBUG_FN( FLG ) if( (debug_level & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
# define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }
#else /* DEBUG */
# define DEBUG_ARGS0( FLG, a0 ) do {} while( 0 )
# define DEBUG_ARGS1( FLG, a0, a1 ) do {} while( 0 )
# define DEBUG_ARGS2( FLG, a0, a1, a2 ) do {} while( 0 )
# define DEBUG_ARGS3( FLG, a0, a1, a2, a3 ) do {} while( 0 )
# define DEBUG_ARGS4( FLG, a0, a1, a2, a3, a4 ) do {} while( 0 )
# define DEBUG_FN( n ) do {} while( 0 )
# define ASSERT( expr, func ) do {} while( 0 )
#endif /* DEBUG */

/* set to nonzero to disable PCI bus checking */
static int disable_pci = 0;

/* ==== note about address mapping ====
 *
 * It is possible to map PCI <--> NS9750 addresses in the context of
 * the mapping between NS9750 physical and kernel virtual addresses,
 * resulting in an identity mapping between kernel virtual address
 * and PCI physical address (both from PCI to NS9750 and from Ns9750
 * to PCI).  This is possible because the NS9750, as an embedded system,
 * has limited RAM capacity.  Address space is not very constrained in
 * this context.
 *
 * In the NS9750:
 * PCI memory base is 0x80000000 (phys), 0xF8000000 (virt)
 * RAM memory base is 0x00000000 (phys), 0xC0000000 (virt)
 *
 * In PCI addressing:
 * PCI memory base 0xF8000000
 * NS9750 base 0xC0000000
 *
 * WARNING WARNING:  These mappings must be revisited when the
 * NS9750 is configured with more than a single bank of SDRAM.
 *
 * PALT0VAL is 0x7C, which is the high order 7 bits of 0xF8000000
 * BAR0 base on bridge is 0xC0000000
 */

#ifdef CONFIG_PCI_PROC_ERRORS
/*
 * Define an array for counting PCI bus error interrupts.
 * One array element for each condition.  The condition
 * bits are shifted left 16 bits because they are expressed
 * in pci.h as if they were in a 16-bit status register. They
 * are in the high-order half of the cmd/sts 32-bit register.
 */
struct pci_error_count
{
	int error_bit;
	unsigned long error_count;
	const char* error_name;
}
pci_error_counters[] =
{
	{PCI_STATUS_PARITY << 16,            0, "MasterParityError"},
	{PCI_STATUS_SIG_TARGET_ABORT << 16,  0, "   SigTargetAbort"},
	{PCI_STATUS_REC_TARGET_ABORT << 16,  0, "   RcvTargetAbort"},
	{PCI_STATUS_REC_MASTER_ABORT << 16,  0, "   RcvMasterAbort"},
	{PCI_STATUS_SIG_SYSTEM_ERROR << 16,  0, "   SigSystemError"},
	{PCI_STATUS_DETECTED_PARITY << 16,   0, "   DetParityError"},
	{0,                            0, 0}
};
#endif /* CONFIG_PCI_PROC_ERRORS */

/*
 * Prepare initial PCI bridge address decoding and mapping.  This
 * function follows most of the pattern described in the
 * NS9750 Sample Driver Configurations document provided by
 * NetSilicon.  The version I am using is 90000574_A, dated
 * March 2004.
 *
 * This function modifies the address decoding from the original
 * description in the DriverConfig document, informed by the
 * "identity mapping" described in earlier commentary.
 */
static void __init ns9750_preinit( void )
{
	DEBUG_ARGS0( DEBUG_INIT, "pci preinit start\n");

	/* Set up bridge configuration in CMD register */
	writel( 0x80000004, NS_PCI_CONFIG_ADDR );	// select CMD reg
	writel( 0x00000046, NS_PCI_CONFIG_DATA );	// init bridge

	/* config BAR0: 0x00000000, 256 Mbytes */
	writel( 0x80000010, NS_PCI_CONFIG_ADDR );	// select BAR0
	writel( 0x0000000, NS_PCI_CONFIG_DATA );	// write base address

	/* Set up I/O space translation */
	writel( 0x0, NS_PCI_PALTIO );

	/* Set up memory space translation */
	writel( 0x43424140, NS_PCI_PALTMEM0 );

	/* Enable AHB-to-PCI translations */
	writel( 0x1, NS_PCI_PALTCTL );

	/* config latency and cache size registers */
	writel( 0x8000000C, NS_PCI_CONFIG_ADDR );	// reg03
	writel( 0x0000FF00, NS_PCI_CONFIG_DATA );	// 255 cycle latency

#if defined(DEBUG_NSPCI)
        {
                static void show_pci_dev(int);
                show_pci_dev(5);
        }
#endif

	DEBUG_ARGS0( DEBUG_INIT, "PCI bridge pre-init complete\n");
}

#if defined(DEBUG_NSPCI)
static void __init show_pci_dev( int devno )
{
	unsigned long config_addr = ((devno & 0x1F) << 11) | 0x80000000;
	int bar;
        unsigned long ldata;

        writel( config_addr + PCI_VENDOR_ID, NS_PCI_CONFIG_ADDR );
        ldata = readl( NS_PCI_CONFIG_DATA );
	DEBUG_ARGS2( DEBUG_INIT, "PCI dev %d: 0x%lx\n", devno, ldata );
	if( ldata == 0xFFFFFFFFUL )
	{
		return;
	}
	/* Read the BARs */
	for(bar = 0; bar < 6;)
	{
		writel( config_addr + PCI_BASE_ADDRESS_0 + (bar << 2), NS_PCI_CONFIG_ADDR );
		ldata = readl( NS_PCI_CONFIG_DATA );
		if( (ldata & (PCI_BASE_ADDRESS_SPACE | PCI_BASE_ADDRESS_MEM_TYPE_MASK)) == PCI_BASE_ADDRESS_MEM_TYPE_64 )
		{
			/* memory with 64-bit address */
			bar += 2;
			DEBUG_ARGS1( DEBUG_INIT, "         64-bit address: %08lx\n", ldata);
			continue;
		}
		if( (ldata & PCI_BASE_ADDRESS_SPACE_IO) && ((ldata & PCI_BASE_ADDRESS_IO_MASK) !=0))
		{
			/* I/O requested */
			DEBUG_ARGS1( DEBUG_INIT, "         I/O: 0x%08lx\n", ldata);
		}
		if( !(ldata & PCI_BASE_ADDRESS_SPACE_IO) && ((ldata & PCI_BASE_ADDRESS_MEM_MASK) !=0))
		{
			/* Memory requested */
			DEBUG_ARGS1( DEBUG_INIT, "         Mem: 0x%08lx\n", ldata);
		}
		bar += 1;
	}
}
#endif

#ifdef CONFIG_PCI_PROC_ERRORS
/***********************************************************************
 * @Function: countPciError
 * @Return: void
 * @Descr: count PCI bus errors for subsequent reporting
 * in /proc/driver/pci_errors
 ***********************************************************************/
static void
countPciError( unsigned long intStat )
{
	struct pci_error_count *counter;
	for(counter = pci_error_counters; counter->error_bit; counter++)
	{
		if( intStat & counter->error_bit )
		{
			DEBUG_ARGS1( DEBUG_VERBOSE, "PCI error %s\n",
					counter->error_name);
			counter->error_count += 1;
		}
	}
}

static int procPciError(char *page, char **start, off_t off,
			int count, int *eof, void *data)
{
	int len = 0;
	char *p = page;
	struct pci_error_count *counter = pci_error_counters;
	for(counter = pci_error_counters; counter->error_bit; counter++)
	{
		if( (p - page) + 70 >= PAGE_SIZE )
		{
			break;
		}
		p += sprintf( p, "%s: %ld\n", counter->error_name, counter->error_count );
	}
	len = p - page;
	if(len < off + count)
	{
		*eof = 1;
	}
	*start = page + off;
	len -= off;
	if(len > count)
	{
		len = count;
	}
	if( len < 0)
	{
		len = 0;
	}
	return len;
}

#endif /* CONFIG_PCI_PROC_ERRORS */

/***********************************************************************
 * @Function: pciHandler
 * @Return: IRQ_HANDLED
 * @Descr: minimal interrupt handler, setting IRQ ACK asap.
 ***********************************************************************/
static irqreturn_t pciHandler( /*@unused@*/ int iIrq,
                               /*@unused@*/ void* pvDevId,
                               /*@unused@*/ struct pt_regs* pRegs )
{
	int errcnt;
	unsigned long intstat;

	errcnt = 3;
	do
	{
		void* errAhbAddr;
		void* errPciAddr;
		unsigned long ahberr;

		ahberr = readl( NS_PCI_PINTR ) & 1 ;
		// writel( 0, NS_PCI_PINTR );
		writel( 1, NS_PCI_PINTR );

		/* Read bridge status */
		writel( 0x80000004, NS_PCI_CONFIG_ADDR );	// select CMD/STS reg
		intstat = readl( NS_PCI_CONFIG_DATA );	// get it
		// clear interrupt sources
		writel( intstat, NS_PCI_CONFIG_DATA );

		errAhbAddr = (void*)readl( NS_PCI_PAHBERR );
		errPciAddr = (void*)readl( NS_PCI_PCIERR );
#ifdef CONFIG_PCI_PROC_ERRORS
		countPciError(intstat);
#endif
		errcnt -= 1;
		DEBUG_ARGS4( DEBUG_VERBOSE, "PCI bus error: PINTR=%lx, status=%lx, pci=%p, ahb=%p\n",
			ahberr, intstat, errPciAddr, errAhbAddr );
		DEBUG_ARGS1( DEBUG_VERBOSE, "ERrPC=0x%lx\n", instruction_pointer(pRegs));
                intstat = 0;
	}
	while( intstat && errcnt );
	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: arbHandler
 * @Return: IRQ_HANDLED
 * @Descr: minimal interrupt handler, setting IRQ ACK asap.
 ***********************************************************************/
static irqreturn_t arbHandler( /*@unused@*/ int iIrq,
                               /*@unused@*/ void* pvDevId,
                               /*@unused@*/ struct pt_regs* pRegs )
{
	int errcnt;
	unsigned long intstat;

	errcnt = 3;
	do
	{
		intstat = readl( NS_PCI_PARBINT ) ;
		writel( intstat, NS_PCI_PARBINT ) ;
		if( intstat & 2 )
		{
			/* Re-enable the broken master */
			unsigned long master_enb = readl( NS_PCI_PARBCFG ) & ~2UL;
			writel( master_enb, NS_PCI_PARBCFG );
			writel( master_enb | 2, NS_PCI_PARBCFG );
		}
		errcnt -= 1;

		// DEBUG_ARGS1( DEBUG_VERBOSE, "PCI arbiter error: PARBINT=%lx\n", intstat );

	}
	while( intstat && errcnt );
	return IRQ_HANDLED;
}

static void __init ns9750_postinit( void )
{
	unsigned long intstat ;
	int errcnt;

	DEBUG_ARGS0( DEBUG_INIT, "pci postinit\n");

	writel( 0x0, NS_PCI_PARBCFG );	// disable external masters
	writel( 0xE, NS_PCI_PARBCFG );	// re-enable external masters

	errcnt = 3;
	do
	{
		intstat = readl( NS_PCI_PARBINT ) ;
		writel( intstat, NS_PCI_PARBINT ) ;
		errcnt -= 1;
		DEBUG_ARGS1( DEBUG_INIT, "PARBINT=%lx\n", intstat );
	}
	while( intstat && errcnt );

	if ((errcnt=request_irq(IRQ_PCI_ARBITER, arbHandler,
				SA_INTERRUPT,"PCIarbiter",NULL))) {
		DEBUG_ARGS1( DEBUG_INIT, "IRQ_PCI_ARBITER %i not available\n", IRQ_PCI_ARBITER );
	}

	if ((errcnt=request_irq(IRQ_PCI_BRIDGE, pciHandler,
				SA_INTERRUPT,"PCIbridge",NULL))) {
		DEBUG_ARGS1( DEBUG_INIT, "IRQ_PCI_BRIDGE %i not available\n", IRQ_PCI_BRIDGE );
	}
	writel( 0xF901, NS_PCI_PINTEN );
	writel( 0x13, NS_PCI_PARBINTEN );
}

static int ns9750_cfg_read(struct pci_bus *bus,
			   unsigned int devfn,
			   int where,
			   int size,
			   u32* val)
{
	unsigned long ldata;
	unsigned long config_addr = (devfn << 8) | 0x80000000;
	writel( config_addr | (where & 0xFC), NS_PCI_CONFIG_ADDR );
	ldata = readl( NS_PCI_CONFIG_DATA );

	DEBUG_ARGS3( DEBUG_INIT, "cfg-read 0x%08lx from %d:%02x\n",
		ldata, devfn >> 3, where >> 2 );

	switch(size)
	{
	case 1:
		ldata >>= (where & 3) << 3;
		*val = ldata & 0xFF;
		break;
	case 2:
		ldata >>= (where & 1) << 4;
		*val = ldata & 0xFFFF;
		break;
	case 4:
		*val = ldata;
		break;
	}
	return PCIBIOS_SUCCESSFUL;
}


static int ns9750_cfg_write(struct pci_bus *bus,
			   unsigned int devfn,
			   int where,
			   int size,
			   u32 val)
{
	unsigned long ldata;
	int shift;
	unsigned long config_addr = (devfn << 8) | 0x80000000;

	writel( config_addr | (where & 0xFC), NS_PCI_CONFIG_ADDR );
	switch(size)
	{
	case 1:
		val &= 0xFF;
		shift = (where & 3) << 3;
		ldata = readl( NS_PCI_CONFIG_DATA );
		ldata = (ldata & ~(0xFF << shift)) | (val << shift);
		break;
	case 2:
		val &= 0xFFFF;
		shift = (where & 2) << 3;
		ldata = readl( NS_PCI_CONFIG_DATA );
		ldata = (ldata & ~(0xFFFF << shift)) | (val << shift);
		break;
	case 4:
		ldata = val;
		break;
	default:
		ldata = readl( NS_PCI_CONFIG_DATA );
		break;
	}
	writel( ldata, NS_PCI_CONFIG_DATA );

	DEBUG_ARGS3( DEBUG_INIT, "cfg-write 0x%08lx to %d:%02x\n",
		ldata, devfn >> 3, where >> 2 );

	return PCIBIOS_SUCCESSFUL;
}


struct pci_bus *sys_bus;
struct pci_ops ns9750_pci_ops =
{
	.read = ns9750_cfg_read,
	.write = ns9750_cfg_write
};

int __init ns9750_setup(int nr, struct pci_sys_data *sys)
{
	struct resource *res;
	int retval;

	if (nr)
		return 0;

	res = kmalloc(sizeof(struct resource) * 2, GFP_KERNEL);
	if (!res) {
		printk("NS9750_PCI: out of memory for root PCI bus resources");
		return 0;
	}

	memset(res, 0, sizeof(struct resource) * 2);

	res[0].flags = IORESOURCE_MEM;
	res[0].name  = "PCI non-prefetch";
	res[1].flags = IORESOURCE_MEM | IORESOURCE_PREFETCH;
	res[1].name  = "PCI prefetch";

	retval = allocate_resource(&iomem_resource, &res[0], 0x00800000,
			  0x80000000, 0x80FFFFFF, 0x100000, NULL, NULL);
	retval = allocate_resource(&iomem_resource, &res[1], 0x00800000,
			  0x80000000, 0x80FFFFFF, 0x100000, NULL, NULL);

	sys->resource[0] = &ioport_resource;
	sys->resource[1] = &res[0];
	sys->resource[2] = &res[1];
	sys->mem_offset  = 0x000000;
	sys->io_offset = 0x000000;

	return 1;
}


struct pci_bus * __init ns9750_scan_bus(int nr, struct pci_sys_data *sys)
{
	return pci_scan_bus(0, &ns9750_pci_ops, sys);
}

/*
 * We now use the slot ID instead of the device identifiers to select
 * which interrupt is routed where.
 */
static int __init ns9750_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	switch (slot) {
	case 0:  /* host bridge */
		return 0;

	case 5:  /* slot */
		printk( KERN_INFO "PCI: Device %s, slot/pin %d/%d, int %d\n",
			pci_name(dev), slot, pin, IRQ_PCI_EXT_0 );
		return IRQ_PCI_EXT_0;

	default:
		printk(KERN_ERR "PCI: unknown device %s, slot/pin %d/%d\n",
			pci_name(dev), slot, pin);
		return 0;
	}
}

static struct hw_pci ns9750_pci __initdata = {
	.swizzle		= pci_std_swizzle,
	.map_irq		= ns9750_map_irq,
	.nr_controllers		= 1,
	.setup			= ns9750_setup,
	.scan			= ns9750_scan_bus,
	.preinit		= ns9750_preinit,
	.postinit		= ns9750_postinit,
};

static int __init ns9750_pci_init(void)
{
	if( disable_pci )
	{
		printk( KERN_WARNING "PCI bus scanning disabled\n");
		return 0;
	}
#ifdef CONFIG_PCI_PROC_ERRORS
	if( !create_proc_read_entry("driver/pci_errors", 0, NULL, procPciError, NULL) )
	{
		printk( KERN_WARNING "Unable to create /proc/driver/pci_errors report file\n");
	}
#endif
        pci_common_init(&ns9750_pci);

	/* Enabling BAR0 is required to force the bridge to
	 * decode 256 Mbytes of PCI space.  BAR0 is zero, in
	 * order to allow the decoded addresses to be at address
	 * zero in AHB address space (this is where RAM is)
	 */
	writel( NS_PCI_EN_BAR0, NS_PCI_PMISC );

	/* config BAR0: 0x00000000, 256 Mbytes */
	writel( 0x80000010, NS_PCI_CONFIG_ADDR );	// select BAR0
	writel( 0x0000000, NS_PCI_CONFIG_DATA );	// write base address

#if defined(DEBUG_NSPCI)
	if( debug_level & DEBUG_INIT )
	{
		show_pci_dev( 0 );
		show_pci_dev( 5 );
	}
#endif
	return 0;
}


static int __devinit disable(char *str)
{
    if( strstr(str, "true") )
    {
        disable_pci = 1;
    }
    return 1;
}

__setup("nopci=", disable);


subsys_initcall(ns9750_pci_init);

