/************************************************************************
 *
 * Copyright (C) 2005 by FS Forth-Systeme GmbH && Sistemas Embebidos S.A.
 * Digi International Companies.
 * All rights reserved.
 *
 * $Id: a9m9xx0_dev-ide.c,v 1.2 2005-09-06 07:34:59 uzeisberger Exp $
 * @Author: Pedro Perez de Heredia
 * @Descr: IDE driver for the CompactFlash inteface for the A9M9750 and
 *         A9M9360 modules. Based on bast-ide.c
 * @References: [1] Logic Specification for A9M9750DEV.
 *                  Version 2.0.
 * @TODO:  * Activate machine checking.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ************************************************************************/

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <linux/delay.h>

#include <asm/mach-types.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arch/a9m9xx0dev.h>	/* FPGA_CF_CONF */
#include <asm/arch/irqs.h>		/* IRQ_FPGA_CF */

// Driver constants
#define	A9M9XX0_MAX_IDE_HWIF		1
#define CF_SHIFT_ADDR			0
#define	A9M9XX0_LOGIC_MINIMUM		0x20
#define	DRV_NAME			"A9M9XX0_dev-ide"
#define	CF_HWIF_OFFSET			FPGA_CF_SIZE

// CF configuration register flags
#define CF_CD_MASK			0x0001
#define CF_RST_MASK			0x0004

// Common Memory/IO internal registers for Storage CF+ cards
#define CF_DATA_OFFSET			(0x00 << CF_SHIFT_ADDR)	// Data
#define CF_ERROR_OFFSET			(0x01 << CF_SHIFT_ADDR)	// Error
#define CF_NSECTOR_OFFSET		(0x02 << CF_SHIFT_ADDR)	// Number of sectors
#define CF_SECTOR_OFFSET		(0x03 << CF_SHIFT_ADDR)	// Starting sector
#define CF_LCYL_OFFSET			(0x04 << CF_SHIFT_ADDR)	// Low Cylinder
#define CF_HCYL_OFFSET			(0x05 << CF_SHIFT_ADDR)	// High Cylinder
#define CF_SELECT_OFFSET		(0x06 << CF_SHIFT_ADDR)	// Select
#define CF_STATUS_OFFSET		(0x07 << CF_SHIFT_ADDR)	// Status
#define CF_CTRL_OFFSET			(0x0e << CF_SHIFT_ADDR)	// Control
#define CF_DATA_DUP_OFFSET		(0x08 << CF_SHIFT_ADDR)	// Duplicate Data Register

#define CF_CONFIG_OPTION_REG_OF		0x200	// Compact Flash Storage card Configuration
						// Option register


/* list of registered interfaces */
static ide_hwif_t *hwifs[A9M9XX0_MAX_IDE_HWIF];

static int __init
a9m9xx0_ide_register( unsigned int base, unsigned int aux, int irq,
		 ide_hwif_t **hwif)
{
	hw_regs_t hw;

	memset( &hw, 0, sizeof(hw) );

	// Configure addresses of IO registers
	// Use duplicate data register instead of data register
	hw.io_ports[IDE_ERROR_OFFSET] = (unsigned long)( base + CF_ERROR_OFFSET );		// Error
	hw.io_ports[IDE_NSECTOR_OFFSET] = (unsigned long)( base + CF_NSECTOR_OFFSET );		// Sector Count
	hw.io_ports[IDE_SECTOR_OFFSET] = (unsigned long)( base + CF_SECTOR_OFFSET );		// Sector No
	hw.io_ports[IDE_LCYL_OFFSET] = (unsigned long)( base + CF_LCYL_OFFSET );		// Low Cylinder
	hw.io_ports[IDE_HCYL_OFFSET] = (unsigned long)( base + CF_HCYL_OFFSET );		// High Cylinder
	hw.io_ports[IDE_SELECT_OFFSET] = (unsigned long)( base + CF_SELECT_OFFSET );		// Select
	hw.io_ports[IDE_STATUS_OFFSET] = (unsigned long)( base + CF_STATUS_OFFSET );		// Status
	hw.io_ports[IDE_CONTROL_OFFSET] = (unsigned long)( base + CF_CTRL_OFFSET );		// Control

	// For the data register, access over the 16bit IO space
	hw.io_ports[IDE_DATA_OFFSET] = (unsigned long)( aux + CF_DATA_DUP_OFFSET );		// Data

	hw.irq = irq;

	ide_register_hw( &hw, hwif );

	return 0;
}

static int __init a9m9xx0_ide_init(void)
{
	unsigned short ctrl_reg;
	unsigned char logic_rev;
	int j;

	// TODO activate following
	//if( !(machine_is_a9m9750() || machine_is_a9m9360()) )
	//    return 0;

	logic_rev = readb( FPGA_LVR2 );
	printk( DRV_NAME " Logic Revison: %d.%d\n", logic_rev >> 4, logic_rev & 0x0f );

	// Check if logic revision supports CF interface
	if( logic_rev < A9M9XX0_LOGIC_MINIMUM ) {
		printk( DRV_NAME " Logic Revison doesn't support CF interface\n" );
		return 0;
	}

	for( j = 0; j < A9M9XX0_MAX_IDE_HWIF; j++ ) {

		// Detect if a CF is inserted on the socket
		ctrl_reg = readb( FPGA_CF_CONF + CF_HWIF_OFFSET * j );

		if( ( ctrl_reg & CF_CD_MASK ) != CF_CD_MASK ) {
			printk( KERN_NOTICE DRV_NAME " CF detected\n" );

			// Reset CF
			writeb( CF_RST_MASK, FPGA_CF_CONF + CF_HWIF_OFFSET * j );
			mdelay(10);
			writeb( 0x00, FPGA_CF_CONF + CF_HWIF_OFFSET * j );
			mdelay(10); // @TODO review following delay...

			// Set CF in IO mode and activate CF IRQ to work in level mode
			writeb( 0x41, FPGA_CF_ATR + CF_CONFIG_OPTION_REG_OF + CF_HWIF_OFFSET * j );

			ctrl_reg = readb( FPGA_CF_ATR + CF_CONFIG_OPTION_REG_OF + CF_HWIF_OFFSET * j );

			//Check if the value writen can be read (some slow CF doesn't work...)
			if( ctrl_reg != 0x41 ) {
				printk( KERN_WARNING DRV_NAME
				": unable to access to Configuration Option Register, 0x%02x\n", ctrl_reg );
				continue;
	        	}

			a9m9xx0_ide_register( FPGA_CF_IO8 + CF_HWIF_OFFSET * j,
	                                      FPGA_CF_IO16 + CF_HWIF_OFFSET * j,
	                                      IRQ_FPGA_CF, &hwifs[j] );
		}
	}

	return 0;
}

module_init( a9m9xx0_ide_init );

MODULE_AUTHOR( "Pedro Perez de Heredia <pperez@embebidos.com>" );
MODULE_LICENSE( "GPL" );
MODULE_DESCRIPTION( "A9M9XX0_DEV IDE driver");
