/************************************************************************
 *
 * Copyright (C) 2005 by Sistemas Embebidos S.A.
 * A Digi International Company.
 * All rights reserved.
 *
 * $Id: a9m9xx0_ts.c,v 1.3 2006-01-02 14:06:51 uzeisberger Exp $
 * @Author: Pedro Perez de Heredia
 * @Descr: Touch screen driver for the...
 * @References: [1] NS9360 Hardware Reference Manual Version 90000675_A
 *                  March 2005.
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 ************************************************************************
 * @History:
 *	14/07/2005	PPH: Initial Version.
 *	29/07/2005	PPH: Improved support for pen-up detection and
 *                           fixed initialization of the input device.
 *
 ************************************************************************
 * @TODO, @TOIMPROVE
 *
 ************************************************************************/

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/serio.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm/arch/a9m9xx0dev.h>
#include <asm/arch/a9m9xx0_ts.h>
#include <asm/arch/ns9xx0_spi.h>


#define A9M9XX0_TS_VERSION	0x0101
#define DRV_NAME		"a9m9xx0_ts"
#define SAMPLE_TIMEOUT		20	/* sample every 20ms */
#define MAX_SAMPLES_AVG		10	/* max number of samples to average */
#define CURRENT_X		touch->x[touch->samples_cnt][touch->cnt]
#define CURRENT_Y		touch->y[touch->samples_cnt][touch->cnt]

//#define DEBUG
#ifdef DEBUG
# define DEBUG_INIT		0x0001
# define DEBUG_PENUP		0x0002
# define DEBUG_COORD		0x0004
# define DEBUG_ALL		0xffff

static int debug_level = DEBUG_PENUP|DEBUG_COORD;

# define DEBUG_DRV( FLG, x... )	if( (debug_level & (FLG)) == (FLG) ) printk(x)
#else /* DEBUG */
# define DEBUG_DRV( FLG, x... )
#endif /* DEBUG */

MODULE_AUTHOR("Pedro Perez de Heredia <pperez@embebidos.com>");
MODULE_DESCRIPTION("A9M9XX0 touchscreen driver");
MODULE_LICENSE("GPL");


// Global variables
static char *a9m9xx0ts_name = "a9m9xx0 TouchScreen";

// The touch-screen struct
struct a9m9xx0ts {
	struct input_dev *dev;
	struct timer_list timer;
	int samples_ms;
	int samples_avg;
	int samples_cnt;
	int cnt;
	unsigned short x[MAX_SAMPLES_AVG][3];
	unsigned short y[MAX_SAMPLES_AVG][3];
	int y_top;
	int y_bottom;
	int x_right;
	int x_left;
	ns_spi_dev *spi;
	char phys[32];
};


/***********************************************************************
 * @Function: a9m9xx0ts_get_better_values
 * @Return: Returns the better values.
 * @Descr: Gets, from 3 samples, the 2 better values and returns the
 *         average of this samples.
 ***********************************************************************/
static unsigned short a9m9xx0ts_get_better_values( unsigned short *vals )
{
	int diff12, diff23, diff31;

	diff12 = ( vals[0] > vals[1] ) ? vals[0] - vals[1] : vals[1] - vals[0];
	diff23 = ( vals[1] > vals[2] ) ? vals[1] - vals[2] : vals[2] - vals[1];
	diff31 = ( vals[2] > vals[0] ) ? vals[2] - vals[0] : vals[0] - vals[2];

	if( diff12 < diff23 && diff12 < diff31 )
		return ( vals[0] + vals[1] )/2;
	if( diff23 < diff12 && diff23 < diff31 )
		return ( vals[1] + vals[2] )/2;

	return ( vals[0] + vals[2] )/2;
}


/***********************************************************************
 * @Function: a9m9xx0ts_get_coordinate
 * @Return: The coordinate value for the selected axis.
 * @Descr: This functions gets the coordinate value, from the Touch
 *         screen controller (using the spi interface) for the selected
 *         axis.
 ***********************************************************************/
static unsigned short a9m9xx0ts_get_coordinate( ns_spi_dev *spi, unsigned int axis )
{
	unsigned int command;
	unsigned int data, temp;

	switch( axis ) {
		case ABS_X:	command = GET_ABS_X;	break;
		case ABS_Y:	command = GET_ABS_Y;	break;
		default:	return 0xffff;
	}

	ns9xx0_spi_tx_dword( spi, command );
	temp = ns9xx0_spi_rx_dword( spi );
	data = (temp & 0x00ff0000) >> 8;
	data |= (temp >> 24);
	data >>= 3;

	return (data & ADS7843_XY_MASK);
}


/***********************************************************************
 * @Function: a9m9xx0ts_process
 * @Return: Nothing.
 * @Descr: Function that reads the x,y coordinates. If invoked from timer
 *         function, enables the touch interrupt line on the FPGA.
 *         Detects when the pen is down, gets the coordinates and disa-
 *         bles the interrupt during a short period of time, using a ker-
 *         nel timer.
 ***********************************************************************/
static void a9m9xx0ts_process( struct input_dev *dev, int timer )
{
	struct a9m9xx0ts *touch = (struct a9m9xx0ts *)dev->private;
	unsigned int data_x = 0, data_y = 0;
	int i;

	if( timer ) {
		// Have been invoked from timer function
		// Enable the touch interrupt to be able to read the status
		// of the pin
		writeb( readb( FPGA_IER ) | IRQ_FPGA_TOUCH_MA, FPGA_IER );
	}

	if( !( readb( FPGA_ISR ) & IRQ_FPGA_TOUCH_MA ) && timer ) {
		// The pen is up...
		input_report_key( dev, BTN_TOUCH, 0 );
 		input_report_abs( dev, ABS_PRESSURE, 0 );
 		input_sync( dev );
		touch->samples_cnt = 0;
		touch->cnt = 0;
		(void)readb( FPGA_ISR ); // Dummy read... something is wrong
		DEBUG_DRV( DEBUG_PENUP, DRV_NAME ": Pen up\n" );
		return;
	}

	// Disable the ts interrupt, will be re-enabled after a brief delay
	writeb( readb( FPGA_IER ) & ~IRQ_FPGA_TOUCH_MA, FPGA_IER );
	mod_timer ( &touch->timer, jiffies + (touch->samples_ms * HZ) / 1000 );

	CURRENT_X = a9m9xx0ts_get_coordinate( touch->spi, ABS_X );
	CURRENT_Y = a9m9xx0ts_get_coordinate( touch->spi, ABS_Y );

	// Validate the value
	if( ( CURRENT_Y > touch->y_bottom && CURRENT_Y < touch->y_top ) &&
	    ( CURRENT_X > touch->x_right && CURRENT_X < touch->x_left ) ) {
		touch->cnt++;
		if( touch->cnt == 3 ) {
			touch->cnt = 0;
			touch->samples_cnt++;
		}
	}

	if( touch->samples_cnt == touch->samples_avg ) {

		touch->samples_cnt = 0;

		for( i = 0; i < touch->samples_avg; i++ ) {
			data_y += a9m9xx0ts_get_better_values( &touch->y[i][0] );
			data_x += a9m9xx0ts_get_better_values( &touch->x[i][0] );
		}
		data_y /= touch->samples_avg;
		data_x /= touch->samples_avg;
		data_y = touch->y_top - data_y;
		data_x = touch->x_left - data_x;

		DEBUG_DRV( DEBUG_COORD, DRV_NAME ": X = %d, Y = %d\n", data_x, data_y );

		input_report_abs( dev, ABS_X, data_x );
		input_report_abs( dev, ABS_Y, data_y );
		input_report_key( dev, BTN_TOUCH, 1 );
		input_report_abs( dev, ABS_PRESSURE, 1 );
		input_sync( dev );
	}
}

/***********************************************************************
 * @Function: a9m9xx0ts_irq_handler
 * @Return: IRQ_HANDLED.
 * @Descr: Interrupt handler for the touch interface. Detects when the
 *         pen is down, gets the coordinates and disables the interrupt
 *         during a short period of time.
 ***********************************************************************/
static irqreturn_t a9m9xx0ts_irq_handler( int irq, void *dev_id, struct pt_regs *regs )
{
	struct input_dev *dev = (struct input_dev *)dev_id;

	a9m9xx0ts_process( dev, 0 );
	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: a9m9xx0ts_timer
 * @Return: nothing
 * @Descr: Invokes the interrupt handler for the Pen-up detection and
 *         enables the ts interrupt.
 ***********************************************************************/
static void a9m9xx0ts_timer( unsigned long val )
{
	struct input_dev *dev = (struct input_dev *)val;

	a9m9xx0ts_process( dev, 1 );
};

/***********************************************************************
 * @Function: a9m9xx0ts_probe
 * @Return: 0 on success, error code otherwise
 * @Descr: The probe method, called at module load time.
 ***********************************************************************/
static int __init a9m9xx0ts_probe( struct device *dev )
{
	struct a9m9xx0ts	*ts;
	struct a9m9xx0_ts_data	*data;
	struct input_dev	*input_dev;
	int			retval;

	DEBUG_DRV( DEBUG_INIT, "a9m9xx0ts_probe\n" );

	if( !dev->platform_data ) {
		printk( KERN_ERR DRV_NAME ": missing platform_data\n" );
		return -ENODEV;
	}

	data = (struct a9m9xx0_ts_data *)dev->platform_data ;

	// Allocate memory
	if( ( ts = kzalloc( sizeof(struct a9m9xx0ts), GFP_KERNEL) ) == NULL ) {
		printk( KERN_ERR DRV_NAME ": unable to get memory for touch screen device\n" );
		return -ENOMEM;
	}
	if (!(ts->dev = input_dev = input_allocate_device())) {
		retval = -ENOMEM;
		goto fail_mem1;
	}

	if( ( ts->spi = kmalloc( sizeof(ns_spi_dev), GFP_KERNEL) ) == NULL ) {

		printk( KERN_ERR DRV_NAME ": unable to get memory for spi device\n" );
		retval = -ENOMEM;
		goto fail_mem2;
	}
	memset( ts->spi, 0, sizeof(ns_spi_dev) );

	dev_set_drvdata( dev, ts );

	// Initialize spi struct with the device settings
	ts->spi->port = data->port;
	ts->spi->master = SPI_MASTER;
	ts->spi->mode = SPI_MODE_0;
	ts->spi->bitrate = data->bitrate;
	ts->spi->bit_order = SPI_MSBF;
	ts->spi->gpios = ns_spi_gpios[ts->spi->port];
	ts->spi->base_addr = addr_ser_channel( NS_SER_CTRL_A, ts->spi->port );
	ts->spi->offset = ( ( ts->spi->port < 2 ) ? 0 : NS_SER_CHAN_OFFSET ) +
	                  ( ( ts->spi->port & 1 ) ? 0 : 0x40 );

	if( ( retval = ns9xx0_spi_request_port( ts->spi ) ) != 0 ) {
		printk( KERN_ERR DRV_NAME ": selected SPI port %d busy\n", ts->spi->port );
		retval = -EBUSY;
		goto fail_mem3;
	}

	// Configure the initial settings
	ns9xx0_spi_port_setup( ts->spi );

	input_dev->evbit[0] = BIT(EV_SYN) | BIT(EV_KEY) | BIT(EV_ABS);
	input_dev->keybit[LONG(BTN_TOUCH)] = BIT(BTN_TOUCH);

	input_set_abs_params( input_dev, ABS_X, 0, data->x_left - data->x_right, 0, 0 );
	input_set_abs_params( input_dev, ABS_Y, 0, data->y_top - data->y_bottom, 0, 0 );
	input_set_abs_params( input_dev, ABS_PRESSURE, 0, 1, 0, 0 );

	sprintf( ts->phys, "ts0" );

	input_dev->private = ts;
	input_dev->name = a9m9xx0ts_name;
	input_dev->phys = ts->phys;
	input_dev->id.bustype = BUS_RS232;
	input_dev->id.vendor = 0xDEAD;
	input_dev->id.product = 0xBEEF;
	input_dev->id.version = A9M9XX0_TS_VERSION;

	ts->samples_ms = data->samples_ms;
	ts->samples_avg = data->samples_avg > MAX_SAMPLES_AVG ?
	                  MAX_SAMPLES_AVG : data->samples_avg;

	ts->y_top = data->y_top;
	ts->y_bottom = data->y_bottom;
	ts->x_right = data->x_right;
	ts->x_left = data->x_left;

	// Initialize the ts timer
	init_timer( &ts->timer );
	ts->timer.function = a9m9xx0ts_timer;
	ts->timer.data = (unsigned long)input_dev;

	if( request_irq( IRQ_FPGA_TOUCH, a9m9xx0ts_irq_handler, SA_SHIRQ,
		"a9m9xx0_ts", ts->dev ) ) {
		printk( KERN_ERR DRV_NAME ": Could not allocate irq: %d\n", IRQ_FPGA_TOUCH );
		retval = -EIO;
		goto fail_irq;
	}

	writeb( readb( FPGA_IER ) | IRQ_FPGA_TOUCH_MA, FPGA_IER );

	printk( KERN_INFO DRV_NAME ": successfully loaded\n" );

	// Register to the input system
	input_register_device( input_dev );

	return 0;

fail_irq:
	ns9xx0_spi_release_port( ts->spi );
fail_mem3:
	kfree( ts->spi );
fail_mem2:
	input_free_device(input_dev);
fail_mem1:
	kfree( ts );

	return retval;
}

/***********************************************************************
 * @Function: a9m9xx0ts_remove
 * @Return: 0 on success, error code otherwise
 * @Descr: The remove method of the driver. Frees the allocated resour-
 *         ces.
 ***********************************************************************/
static int a9m9xx0ts_remove( struct device *dev )
{
	struct a9m9xx0ts	*ts;

	DEBUG_DRV( DEBUG_INIT, "a9m9xx0ts_remove\n" );

	ts = dev_get_drvdata( dev );

	free_irq( IRQ_FPGA_TOUCH, ts->dev );
	del_timer_sync( &ts->timer );
	input_unregister_device( ts->dev );
	ns9xx0_spi_release_port( ts->spi );
	kfree( ts->spi );
	kfree( ts );

	return 0;
}

static struct device_driver a9m9xx0ts_driver = {
       .name           = "a9m9xx0-ts",
       .bus            = &platform_bus_type,
       .probe          = a9m9xx0ts_probe,
       .remove         = a9m9xx0ts_remove,
};


int __init a9m9xx0ts_init( void )
{
	DEBUG_DRV( DEBUG_INIT, "Registering driver: " DRV_NAME "\n" );
	return driver_register( &a9m9xx0ts_driver );
}

void __exit a9m9xx0ts_exit( void )
{
	DEBUG_DRV( DEBUG_INIT, "Unregistering driver: " DRV_NAME "\n" );
	driver_unregister( &a9m9xx0ts_driver );
}

module_init(a9m9xx0ts_init);
module_exit(a9m9xx0ts_exit);

