/* linux/drivers/mtd/nand/ccx9x_nand.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*
 *  !Revision:   $Revision: 1.8 $: 
 *  !Author:     Markus Pietrek
 *  !Descr:      NAND driver for cc[w]9c platforms
 *  !References: [1] drivers/nand/s3c2410.c
 *               [2] Micron MT29F2G08AABWP (2gb_nand_m29b.pdf)
 *               [3] ST NAND01G-B (2kByte-Secotr size nand01gr3b.pdf)
 *               [4] ST NAND128: (512Byte-Sector size.pdf)
 *
 * Read Busy Time:
 *   [2] page 44: 25us
 *   [3] page 43: 25us
 *   [4] page 37: 15us
*/

#include <linux/config.h>          /* system configuration */

#include <linux/module.h>	   /* MODULE_PARAM */
#include <linux/platform_device.h> /* platform_get */

#include <linux/mtd/nand.h>	   /* nand_chip */
#include <linux/mtd/partitions.h>  /* mtd_partition */

#include <asm/io.h>		   /* ioremap */
#include <asm/mach-types.h>	   /* machine_is_cc9c */
#include <asm/arch-ns9xxx/hardware.h>  /* request_gpio */
#include <asm/arch-ns9xxx/bbus.h>  /* set_gpio_cfg */

#define DRIVER_NAME		"ccx9x_nand"
#define DEVICE_NAME		"onboard_boot"

#define PHYS_NAND_FLASH		( 0x50000000 ) /* NAND Flash Bank #1 */
#define MIN_NEEDED_DELAY	( 25 )         /* Read Busy Time, see above */

#define to_info( pdev ) 	platform_get_drvdata( pdev )
#define mtd_to_info( mtd ) 	container_of( mtd, ccx9x_nand_info_t, mtd )

#define CLEAR(x)     		memset( (x), 0, sizeof( *(x) ) )
#define ERROR( format, ... ) 	printk( KERN_ERR \
					" %s: " format "\n",	\
					__func__,		\
					##__VA_ARGS__ )

#define GOTO_ERROR( ERRNO, MSG, ... )		\
	do { \
		ERROR( MSG, ##__VA_ARGS__ );			\
		err = ERRNO; \
		goto error;  \
	} while( 0 )

/* ********** local function definitions  ********** */

static int  ccx9x_nand_probe(  struct platform_device* pdev );
static int  ccx9x_nand_remove( struct platform_device* pdev );
static void ccx9x_nand_release_device( struct device* dev );
static void ccx9x_nand_hwcontrol( struct mtd_info* mtd, int cmd );
static int  ccx9x_nand_dev_ready( struct mtd_info* mtd );
static int  __init ccx9x_nand_init( void );
static void __exit ccx9x_nand_exit( void );

/* ********** local variables ********** */

/**
 * struct ccx9x_nand_info - Describes the NAND device supported by this driver
 * @mtd: MTD Device
 * @chip: physical chip information
 * @vbase: virtual base address of mapped memory
 * @vdata: virtual address for read/write operations
 * @vcmd: virtual address for command operations
 * @vaddr: virtual address for address operations
 */
typedef struct {
	struct mtd_info  mtd;
	struct nand_chip chip;

	void __iomem*    vbase;
	void __iomem*    vdata;
	void __iomem*    vcmd;
	void __iomem*    vaddr;

        int gpio;
} ccx9x_nand_info_t;

/* define the resources the driver will use */
static struct resource ccx9x_nand_resource = {
	.name  = DRIVER_NAME,
	.start = PHYS_NAND_FLASH,
	.end   = PHYS_NAND_FLASH + SZ_1M,
	.flags = IORESOURCE_MEM
};

/* describes the device */
static struct platform_device ccx9x_nand_device = {
	.id     = -1,
	.name  = DRIVER_NAME, /* must be equal to platform-driver.driver.name*/
	.resource = &ccx9x_nand_resource,
	.dev = {
		.release = ccx9x_nand_release_device,
	},
};

/* describes the driver */
static struct platform_driver ccx9x_nand_driver = {
	.probe  = ccx9x_nand_probe,
	.remove = ccx9x_nand_remove,
	.driver = {
		.name  = DRIVER_NAME,
		.owner = THIS_MODULE,
	},
};

/* our version */
static const char* ccx9x_nand_version =
    "CC9X NAND Driver $Revision: 1.8 $, (c) 2006 Digi International GmbH\n";

#ifdef CONFIG_MTD_PARTITIONS
static const char* probes[] = { "cmdlinepart", NULL };
#endif /* CONFIG_MTD_PARTITIONS */

/* ********** Local Function Implementations  ********** */

/**
 * ccx9x_nand_probe - probes for the device
 * @pdev: device to probe for
 *
 * works on cc9p, cc9c and ccw9c systems
 */
static int ccx9x_nand_probe( struct platform_device* pdev )
{
	struct mtd_partition* mtd_parts = 0;
	ccx9x_nand_info_t*    info      = NULL;
	const char*           part_type = "";
	int mtd_parts_nb = 0;
	int err          = 0;
	
	if( ! (	machine_is_cc9c()        || machine_is_ccw9c() ||
		machine_is_cc9p9360dev() || machine_is_cc9p9360val() ||
		machine_is_cc9p9750dev() || machine_is_cc9p9750val() ||
		machine_is_cc9p9360js() ) )
		GOTO_ERROR( -ENODEV, "System not supported" );

	info = kmalloc( sizeof( *info ), GFP_KERNEL );
	if( NULL == info )
		GOTO_ERROR( -ENOMEM, "No Memory for info" );

	CLEAR( info );
	platform_set_drvdata( pdev, info );
        info->gpio = -1;        /* GPIO not present */
        
	/* request resources and map memory */
	if( request_resource( &iomem_resource, pdev->resource ) )
		GOTO_ERROR( -EBUSY,
			    "Memory already in used: 0x%08lx...0x%08lx",
			    pdev->resource->start,
			    pdev->resource->end );
		
	info->vbase = ioremap( pdev->resource->start,
			       pdev->resource->end - pdev->resource->start );
	if( NULL == info->vbase )
		GOTO_ERROR( -ENOMEM, "ioremap failed" );

	dev_dbg( &pdev->dev, "NAND Flash memory mapped to virtual %p\n",
                 info->vbase );

	/* determine addresses for command and address cycles */
	if( machine_is_ccw9c() || machine_is_cc9c() ) {
		info->vdata = info->vbase + 0x00000000;
		info->vaddr = info->vbase + 0x00008000;
		info->vcmd  = info->vbase + 0x00010000;
	} else if( machine_is_cc9p9360dev() ||  machine_is_cc9p9360val() ||
                   machine_is_cc9p9750dev() ||  machine_is_cc9p9750val() ||
                   machine_is_cc9p9360js() ) {
		info->vdata = info->vbase + 0x00000000;
		info->vaddr = info->vbase + 0x00002000;
		info->vcmd  = info->vbase + 0x00004000;

                info->gpio  = 49;
	} else
		GOTO_ERROR( -ENODEV, "address setup is incomplete. Please add platform here: %s", __FILE__ );

	/* setup device info */
	info->mtd.name = DEVICE_NAME;
	info->mtd.priv = &info->chip;

	info->chip.IO_ADDR_R  = info->vdata;
	info->chip.IO_ADDR_W  = info->vdata;
	info->chip.hwcontrol  = ccx9x_nand_hwcontrol;
	info->chip.eccmode    = NAND_ECC_SOFT;
	info->chip.chip_delay = MIN_NEEDED_DELAY;  /* !see references */
	info->chip.controller = NULL;  /* we have no controller accessing
					  multiple devices */

        if( -1 != info->gpio ) {
                if( !ns9xxx_request_gpio( info->gpio, DRIVER_NAME ) ) {
                        info->chip.dev_ready = ccx9x_nand_dev_ready;
                        ns9xxx_gpio_setpin( info->gpio, NS_GPIO_CFG_INPUT | NS_GPIO_CFG_FUNC_0 );
                } else {
                        /* no need to not work. We are just a slower */ 
                        printk( KERN_WARNING DRIVER_NAME " : Couldn't register GPIO %i, using fixed-delay mode", info->gpio );
                        info->gpio = -1;
                }
        } /* if( -1 != info->gpio */
        
	if( nand_scan( &info->mtd, 1 ) )
		GOTO_ERROR( -ENODEV, "nand_scan failed" );

	/* read partition information */
#ifdef CONFIG_MTD_CMDLINE_PARTS
	mtd_parts_nb = parse_mtd_partitions( &info->mtd, probes, &mtd_parts,0);
	if( mtd_parts_nb > 0 )
		part_type = "command line";
	else
		mtd_parts_nb = 0;
#endif	/* CONFIG_MTD_CMDLINE_PARTS */

	/* Register the partitions */
	dev_notice( &pdev->dev, "Using %s partition definition\n", part_type );
	add_mtd_partitions( &info->mtd, mtd_parts, mtd_parts_nb );
		
	return 0;

error:
	ccx9x_nand_remove( pdev );
	
	return err;
}

/**
 * ccx9x_nand_remove - removes the device
 * @pdev: device to remove
 */
static int ccx9x_nand_remove( struct platform_device* pdev )
{
	ccx9x_nand_info_t* info = to_info( pdev );
	
	if( NULL == info )
		return 0;

        if( -1 != info->gpio )
                ns9xxx_release_gpio( info->gpio );
        
	if( NULL != info->vbase ) {
		iounmap( info->vbase );
		info->vbase = NULL;
	}
	
	release_resource( pdev->resource );

	kfree( info );
	info = NULL;
	
	return 0;
}

/**
 * ccx9x_nand_release_device
 * @dev: device to remove
 */
static void ccx9x_nand_release_device( struct device* dev )
{
	/* nothing to do. But we need a !NULL function */
}

/**
 * ccx9x_nand_hwcontrol - determines the address to use for the operation
 * @mtd: mtd to work on
 * @cmd: command to execute
 */
static void ccx9x_nand_hwcontrol( struct mtd_info* mtd, int cmd )
{
	ccx9x_nand_info_t* info = mtd_to_info( mtd );
	struct nand_chip * chip = mtd->priv;

	switch( cmd ) {
	    case NAND_CTL_SETCLE:
		chip->IO_ADDR_W = info->vcmd;
		break;

	    case NAND_CTL_SETALE:
		chip->IO_ADDR_W = info->vaddr;
		break;

	    default:
		chip->IO_ADDR_W = info->vdata;
		break;
	}
}

/**
 * ccx9x_nand_dev_ready - determines whether the chip has completed a command
 * @mtd: mtd to work on
 *
 * @return: 0 if device is busy, otherwise 1
 */
static int ccx9x_nand_dev_ready( struct mtd_info* mtd )
{
	ccx9x_nand_info_t* info = mtd_to_info( mtd );

        return get_gpio_stat( info->gpio );
}
    
/**
 * ccx9x_nand_init - initializes the driver
 */
static int __init ccx9x_nand_init( void )
{
	int err = 0;
	
	printk( KERN_INFO "%s", ccx9x_nand_version );

	err = platform_device_register( &ccx9x_nand_device );
	if( err )
		GOTO_ERROR( -ENODEV, "Device Registering failed" );

	err = platform_driver_register( &ccx9x_nand_driver );
	if( err )
		GOTO_ERROR( -ENODEV, "Driver Registering failed" );

	return 0;

error:
	return err;
}

/**
 * ccx9x_nand_exit - removes the driver
 */
static void __exit ccx9x_nand_exit( void )
{
	platform_driver_unregister( &ccx9x_nand_driver );
	platform_device_unregister( &ccx9x_nand_device );
}

module_init( ccx9x_nand_init );
module_exit( ccx9x_nand_exit );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Markus Pietrek <markus_pietrek@digi.com>" );
MODULE_DESCRIPTION( "CCX9X MTD NAND driver" );
