/* linux/drivers/spi/spi_s3c24xx.c
 *
 * Copyright (c) 2006 Ben Dooks
 * Copyright (c) 2006 Simtec Electronics
 *	Ben Dooks <ben@simtec.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
*/


//#define DEBUG

#include <linux/config.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/spi/spi.h>
#include <linux/spi/spi_bitbang.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/hardware.h>

#include <asm/arch/system.h>
#include <asm/arch/regs-ser.h>
#include <asm/arch/regs-bbus.h>
#include <asm/arch/spi.h>
#include <asm/arch/dma.h>

struct ns9xxx_spi {
	spinlock_t		lock;
	void __iomem		*regs;
	int			irq_rx;
	int			irq_tx;
	int			dma_rx;
	int			dma_tx;
	int			len;
	int			count;

	struct resource		*ioarea;
	struct spi_master	*master;
	struct spi_device	*curdev;
	struct device		*dev;
	struct ns9xxx_spi_cfg	*pdata;
	struct platform_device	*pdev;

	u8			stopping;
	struct list_head	queue;
	struct spi_transfer	*current_transfer;
	unsigned long		remaining_bytes;
	u32			flags;
	unsigned int		dma_xfer_len;
	void			*buf_tx;
	void			*buf_rx;
	dma_addr_t		buf_tx_dma;
	dma_addr_t		buf_rx_dma;
	dma_addr_t		desc_tx_dma;
	dma_addr_t		desc_rx_dma;
	ns_buffer_desc_t	*desc_tx;
	ns_buffer_desc_t	*desc_rx;
};

/* Order is SPI_TX, SPI_RX, SPI_CLK and SPI_EN */
static const char ns9xxx_spi_gpios[][4] = {
	{  0,  1,  6,  7 }, /* Port B */	
	{  8,  9, 14, 15 }, /* Port A */	
	{ 40, 41, 22, 23 }, /* Port C */	
	{ 44, 45, 26, 27 }, /* Port D */		
};

#define	SPI_DMA_TX_CTRL_OPT	(NS_BBUS_DMA_CTRL_MODE_FBR | \
				 NS_BBUS_DMA_CTRL_BTE_1)
#define	SPI_DMA_RX_CTRL_OPT	(NS_BBUS_DMA_CTRL_MODE_FBW | \
				 NS_BBUS_DMA_CTRL_BTE_1)

#define	SPI_DMA_TX_STAT_OPT	(NS_BBUS_DMA_STAT_NCIE | \
				 NS_BBUS_DMA_STAT_ECIE | \
				 NS_BBUS_DMA_STAT_NRIE | \
				 NS_BBUS_DMA_STAT_CAIE | \
				 NS_BBUS_DMA_STAT_PCIE)
#define	SPI_DMA_RX_STAT_OPT	SPI_DMA_TX_STAT_OPT

#define	DRIVER_NAME		"ns9xxx_spi"
#define BUFFER_SIZE		PAGE_SIZE
#define INVALID_DMA_ADDRESS	0xffffffff
#define SPI_RX_IDONE		0x00000001
#define SPI_TX_IDONE		0x00000002
#define	SPI_DEFAULT_BPS		(1000000)
#define	SPI_CS_ACTIVE		1
#define	SPI_CS_INACTIVE		0


#define cpu_readl(port, reg)		__raw_readl((port)->regs + (reg))
#define cpu_writel(port, reg, value)	__raw_writel((value), (port)->regs + (reg))

static inline void ns9xxx_spi_stop_tx_dma(struct ns9xxx_spi *hw)
{
	u32 ctrla;

	ctrla = cpu_readl(hw, SER_CTRLA);
	ctrla &= ~NS_SER_CTRL_A_ETXDMA;
	cpu_writel(hw, SER_CTRLA, ctrla);
}

static inline void ns9xxx_spi_start_tx_dma(struct ns9xxx_spi *hw)
{
	u32 ctrla;

	ctrla = cpu_readl(hw, SER_CTRLA);
	ctrla |= NS_SER_CTRL_A_ETXDMA;
	cpu_writel(hw, SER_CTRLA, ctrla);
}

static inline void ns9xxx_spi_stop_rx_dma(struct ns9xxx_spi *hw)
{
	u32 ctrla;

	ctrla = cpu_readl(hw, SER_CTRLA);
	ctrla &= ~NS_SER_CTRL_A_ERXDMA;
	cpu_writel(hw, SER_CTRLA, ctrla);
}
static inline void ns9xxx_spi_start_rx_dma(struct ns9xxx_spi *hw)
{
	u32 ctrla;

	ctrla = cpu_readl(hw, SER_CTRLA);
	ctrla |= NS_SER_CTRL_A_ERXDMA;
	cpu_writel(hw, SER_CTRLA, ctrla);
}

static void ns9xxx_spi_free_dma(struct ns9xxx_spi *hw)
{
	if (hw->buf_tx)
		dma_free_coherent(&hw->pdev->dev, BUFFER_SIZE, hw->buf_tx,
				  hw->buf_tx_dma);
	if (hw->buf_rx)
		dma_free_coherent(&hw->pdev->dev, BUFFER_SIZE, hw->buf_rx,
				  hw->buf_rx_dma);
	if (hw->desc_rx)
		dma_free_coherent(&hw->pdev->dev, sizeof(ns_buffer_desc_t), hw->desc_rx,
				  hw->desc_rx_dma);
	if (hw->desc_tx)
		dma_free_coherent(&hw->pdev->dev, sizeof(ns_buffer_desc_t), hw->desc_tx,
				  hw->desc_tx_dma);
}

static int ns9xxx_spi_alloc_dma(struct ns9xxx_spi *hw)
{
	int ret = -ENOMEM;

	hw->buf_tx = dma_alloc_coherent(&hw->pdev->dev, BUFFER_SIZE,
					&hw->buf_tx_dma, GFP_KERNEL);
	if (!hw->buf_tx)
		goto err_dma_alloc;

	hw->buf_rx = dma_alloc_coherent(&hw->pdev->dev, BUFFER_SIZE,
					&hw->buf_rx_dma, GFP_KERNEL);
	if (!hw->buf_rx)
		goto err_dma_alloc;

	hw->desc_tx = (ns_buffer_desc_t *)dma_alloc_coherent(&hw->pdev->dev,
							     sizeof(ns_buffer_desc_t),
							     &hw->desc_tx_dma, GFP_DMA);
	if (!hw->desc_tx)
		goto err_dma_alloc;

	hw->desc_rx = (ns_buffer_desc_t *)dma_alloc_coherent(&hw->pdev->dev,
							     sizeof(ns_buffer_desc_t),
							     &hw->desc_rx_dma, GFP_DMA);
	if (!hw->desc_rx)
		goto err_dma_alloc;

	return 0;

err_dma_alloc:
	ns9xxx_spi_free_dma(hw);
	return ret;
}

static int ns9xx0_spi_request_gpios(struct ns9xxx_spi *hw)
{
	int ret = 0, i, port = hw->master->bus_num;
	
	for (i = 0; i < 4; i++) {

		if (i==3 && (hw->pdata->set_cs || hw->pdata->use_gpio_for_cs))
			continue;
		if ((ret = ns9xxx_request_gpio(ns9xxx_spi_gpios[port][i], hw->pdev->name)) != 0)
			goto err_gpio;
		ns9xxx_gpio_cfgpin(ns9xxx_spi_gpios[port][i], GPIO_CFG_FUNC_0);
	}
	return ret;
	
err_gpio:
	for (i--; i > 0; i--)
		ns9xxx_release_gpio(ns9xxx_spi_gpios[port][i]);

	return ret;
}

static void ns9xx0_spi_release_gpios(struct ns9xxx_spi *hw)
{
	int i, port = hw->master->bus_num;
	
	for (i = 0; i < 4; i++) {
		if (i==3 && (hw->pdata->set_cs || hw->pdata->use_gpio_for_cs))
			continue;
		ns9xxx_release_gpio(ns9xxx_spi_gpios[port][i]);
	}
}

static inline void ns9xx0_spi_chipsel(struct spi_device *spi, int cs, int active)
{
	struct ns9xxx_spi *hw = spi_master_get_devdata(spi->master);
	unsigned int cspol = spi->mode & SPI_CS_HIGH ? 1 : 0;

	/* Let see if there is some board specific implementation */
	if (hw->pdata->set_cs) {
		hw->pdata->set_cs(hw->pdata, cs, active, cspol);
	} else if (hw->pdata->use_gpio_for_cs == 1) {
		ns9xxx_gpio_setpin(hw->pdata->gpio_cs, cspol == active);
	}
	/* or CS will be automatically controlled by the processor spi engine */
}


static int ns9xxx_spi_setup(struct spi_device *spi)
{
	unsigned int bpw;
	unsigned int hz;
	unsigned int N_value;
	unsigned int br_reg = 0;
	unsigned int ctrla_reg = 0;
	unsigned int ctrlb_reg = 0;
	struct ns9xxx_spi *hw;
	int ret;

	hw = spi_master_get_devdata(spi->master);

	if (hw->stopping)
		return -ESHUTDOWN;

	if (spi->chip_select > spi->master->num_chipselect) {
		dev_dbg(&spi->dev,
				"setup: invalid chipselect %u (%u defined)\n",
				spi->chip_select, spi->master->num_chipselect);
		return -EINVAL;
	}

	if (hw->pdata->use_gpio_for_cs) {
		if ((ret = ns9xxx_request_gpio(hw->pdata->gpio_cs, hw->pdev->name)) != 0) {
			dev_err(&spi->dev, "Cannot request GPIO %d for SPI CS\n",
				hw->pdata->gpio_cs);
			return ret;
		}
		ns9xxx_gpio_cfgpin(hw->pdata->gpio_cs, GPIO_CFG_OUTPUT);
	}

	bpw = spi->bits_per_word;

	if (!bpw)
		bpw = 8;
	if (bpw != 8)
		return -EINVAL;

	if (spi->mode & SPI_CS_HIGH) {
		/* Invert pin if spi engine controls the spi CS */
		if (!hw->pdata->set_cs && !hw->pdata->use_gpio_for_cs)
			(void)ns9xxx_gpio_invert(ns9xxx_spi_gpios[spi->master->bus_num][3], 1);
	}

	if (spi->max_speed_hz) {
		if (spi->max_speed_hz > ns9xxx_bbusclock() / 8) {
			hz = ns9xxx_bbusclock() / 8;
		} else {
			hz = spi->max_speed_hz;
		}
	} else {
		hz = SPI_DEFAULT_BPS;
		dev_info(&spi->dev,
				"setup: max_speed_hz not provided, using default bps %d)\n",
				SPI_DEFAULT_BPS);
	}
	
	/* Reset previous configuration */
	cpu_writel(hw, SER_CTRLA, ctrla_reg);
	
	/* Configure bitrate register */
	N_value = ns9xxx_bbusclock() / ( hz * 2 ) - 1;
	
	if ((N_value & NS_SER_BITRATE_N_MA) != N_value) {
		dev_err(&spi->dev, "invalid bitrate %d (N_value %d)\n", hz, N_value);
		return -EINVAL;
	}

	br_reg = NS_SER_BITRATE_EBIT | NS_SER_BITRATE_CLKMUX_BCLK | NS_SER_BITRATE_TMODE | N_value;
	
	if ((spi->mode & SPI_MODE_3) == SPI_MODE_0)
		br_reg |= NS_SER_BITRATE_SPCPOL_LOW;
	else if ((spi->mode & SPI_MODE_3) == SPI_MODE_1)
		br_reg |= NS_SER_BITRATE_TXCINV | NS_SER_BITRATE_RXCINV;
	else if ((spi->mode & SPI_MODE_3) == SPI_MODE_2)
		br_reg |= NS_SER_BITRATE_SPCPOL_LOW | NS_SER_BITRATE_TXCINV | NS_SER_BITRATE_RXCINV;
	
	cpu_writel(hw, SER_BITRATE, br_reg);
	
	/* Configure bitorder and Master Operation */
	if (!(spi->mode & SPI_LSB_FIRST))
		ctrlb_reg |= NS_SER_CTRL_B_BITORDR;
	
	ctrlb_reg |= NS_SER_CTRL_B_MODE_SPI_M;
	cpu_writel(hw, SER_CTRLB, ctrlb_reg);

	/* Configure other registers */
	cpu_writel(hw, SER_RXBUFTIMER, 0x0);
	cpu_writel(hw, SER_RXCHARTIMER, 0x0);
	cpu_writel(hw, SER_RXMATCH, 0x0);
	cpu_writel(hw, SER_RXMATCHMASK, 0x0);

	ctrla_reg = NS_SER_CTRL_A_CE | NS_SER_CTRL_A_WLS_8;
	
	/* Enable this channel */
	cpu_writel(hw, SER_CTRLA, ctrla_reg);

	dev_dbg(&spi->dev, "%s: mode %d, %u bpw, %d hz\n",
		__FUNCTION__, spi->mode, spi->bits_per_word,
		spi->max_speed_hz);

	return 0;
}

static void ns9xxx_spi_cleanup(const struct spi_device *spi)
{
	struct ns9xxx_spi *hw = spi_master_get_devdata(spi->master);

	if (hw->pdata->use_gpio_for_cs) {
		ns9xxx_release_gpio(hw->pdata->gpio_cs);
	}
}

static void ns9xxx_spi_dma_map_xfer(struct ns9xxx_spi *hw,
				   struct spi_transfer *xfer)
{
	xfer->tx_dma = xfer->rx_dma = INVALID_DMA_ADDRESS;
	if (!(xfer->len & (L1_CACHE_BYTES - 1))) {
		if (xfer->tx_buf
		    && !((unsigned long)xfer->tx_buf & (L1_CACHE_BYTES - 1)))
			xfer->tx_dma = dma_map_single(&hw->pdev->dev,
						      (void *)xfer->tx_buf,
						      xfer->len,
						      DMA_TO_DEVICE);
		if (xfer->rx_buf
		    && !((unsigned long)xfer->rx_buf & (L1_CACHE_BYTES - 1)))
			xfer->rx_dma = dma_map_single(&hw->pdev->dev,
						      (void *)xfer->rx_buf,
						      xfer->len,
						      DMA_FROM_DEVICE);
	}
}


/*
 * Submit next transfer for DMA.
 * lock is held, spi irq is blocked
 */
static void ns9xxx_spi_next_xfer(struct spi_master *master,
				struct spi_message *msg)
{
	struct ns9xxx_spi	*hw = spi_master_get_devdata(master);
	struct spi_transfer	*xfer;
	u32			len;
	dma_addr_t		tx_dma, rx_dma;

	xfer = hw->current_transfer;
	if (!xfer || hw->remaining_bytes == 0) {
		if (xfer)
			xfer = list_entry(xfer->transfer_list.next,
					  struct spi_transfer, transfer_list);
		else
			xfer = list_entry(msg->transfers.next, struct spi_transfer,
					  transfer_list);
		hw->remaining_bytes = xfer->len;
		hw->current_transfer = xfer;
	}
 
	len = hw->remaining_bytes;
 
	tx_dma = xfer->tx_dma;
	rx_dma = xfer->rx_dma;
 
	if (rx_dma == INVALID_DMA_ADDRESS) {
		rx_dma = hw->buf_rx_dma;
		if (len > BUFFER_SIZE)
			len = BUFFER_SIZE;
	}
	if (tx_dma == INVALID_DMA_ADDRESS) {
		if (xfer->tx_buf) {
			tx_dma = hw->buf_tx_dma;
			if (len > BUFFER_SIZE)
				len = BUFFER_SIZE;
			memcpy(hw->buf_tx, xfer->tx_buf + xfer->len - hw->remaining_bytes, len);
			dma_sync_single_for_device(&hw->pdev->dev,
						   hw->buf_tx_dma, len,
						   DMA_TO_DEVICE);
		} else {
			/* Send undefined data; rx_dma is handy */
			tx_dma = rx_dma;
		}
	}

	/* Setup the DMA channels for the transfer */
	ns9xxx_disable_dma_ch(0, hw->dma_tx);	
	ns9xxx_disable_dma_ch(0, hw->dma_rx);
	
	hw->desc_tx->src = (u32 *)(tx_dma);
	hw->desc_tx->len = len;
	hw->desc_tx->s.bits.status = 0;
	hw->desc_tx->s.bits.wrap = 1;
	hw->desc_tx->s.bits.intr = 1;
	hw->desc_tx->s.bits.full = 1;
	hw->desc_tx->s.bits.last = 1;

	hw->desc_rx->src = (u32 *)(rx_dma);
	hw->desc_rx->len = len;
	hw->desc_rx->s.bits.wrap = 1;
	hw->desc_rx->s.bits.status = 0;
	hw->desc_rx->s.bits.last = 0;
	hw->desc_rx->s.bits.intr = 0;
	hw->desc_rx->s.bits.full = 0;

	hw->dma_xfer_len = len;

	ns9xxx_set_options_dma_ch(0, hw->dma_tx, hw->desc_tx_dma,
				  SPI_DMA_TX_CTRL_OPT, SPI_DMA_TX_STAT_OPT);

	ns9xxx_set_options_dma_ch(0, hw->dma_rx, hw->desc_rx_dma,
				  SPI_DMA_RX_CTRL_OPT, SPI_DMA_RX_STAT_OPT);

	dev_dbg(&msg->spi->dev,
		"start xfer %p: len %u tx %p/0x%08x rx %p/0x%08x\n",
		xfer, xfer->len, xfer->tx_buf, xfer->tx_dma,
		xfer->rx_buf, xfer->rx_dma);

	ns9xxx_enable_dma_ch(0, hw->dma_tx);
	ns9xxx_enable_dma_ch(0, hw->dma_rx);
	ns9xxx_spi_start_rx_dma(hw);
	ns9xxx_spi_start_tx_dma(hw);
}

static void ns9xxx_spi_next_message(struct spi_master *master)
{
	struct ns9xxx_spi	*hw = spi_master_get_devdata(master);
	struct spi_message	*msg;

	BUG_ON(hw->current_transfer);

	msg = list_entry(hw->queue.next, struct spi_message, queue);

	ns9xx0_spi_chipsel(msg->spi, msg->spi->chip_select, SPI_CS_ACTIVE);
	ns9xxx_spi_next_xfer(master, msg);
}


static int ns9xxx_spi_transfer(struct spi_device *spi, struct spi_message *msg)
{
	struct ns9xxx_spi	*hw;
	struct spi_transfer	*xfer;
	unsigned long		flags;
	struct device		*controller = spi->master->cdev.dev;

	hw = spi_master_get_devdata(spi->master);

	dev_dbg(controller, "new message %p submitted for %s\n",
			msg, spi->dev.bus_id);

	if (unlikely(list_empty(&msg->transfers)
			|| !spi->max_speed_hz))
		return -EINVAL;
	if (hw->stopping)
		return -ESHUTDOWN;
 
	list_for_each_entry(xfer, &msg->transfers, transfer_list) {
		if (!(xfer->tx_buf || xfer->rx_buf)) {
			dev_dbg(&spi->dev, "missing rx or tx buf\n");
			return -EINVAL;
		}
 
		/* FIXME implement these protocol options!! */
		if (xfer->bits_per_word || xfer->speed_hz) {
			dev_dbg(&spi->dev, "no protocol options yet\n");
			return -ENOPROTOOPT;
		}
	}
 
	/* scrub dcache "early" */
	if (!msg->is_dma_mapped) {
		list_for_each_entry(xfer, &msg->transfers, transfer_list)
			ns9xxx_spi_dma_map_xfer(hw, xfer);
	}
 
	list_for_each_entry(xfer, &msg->transfers, transfer_list) {
		dev_dbg(controller,
			"  xfer %p: len %u tx %p/%08x rx %p/%08x\n",
			xfer, xfer->len,
			xfer->tx_buf, xfer->tx_dma,
			xfer->rx_buf, xfer->rx_dma);
	}
 
	msg->status = -EINPROGRESS;
	msg->actual_length = 0;

	spin_lock_irqsave(&hw->lock, flags);
	list_add_tail(&msg->queue, &hw->queue);
	if (!hw->current_transfer)
		ns9xxx_spi_next_message(spi->master);
	spin_unlock_irqrestore(&hw->lock, flags);

	return 0;
}

static void ns9xxx_spi_irq_trigger_next(struct spi_master *master)
{
	struct ns9xxx_spi	*hw = spi_master_get_devdata(master);
	struct spi_message	*msg;
	struct spi_transfer	*xfer;
	u32			len;

	BUG_ON((hw->flags & (SPI_TX_IDONE|SPI_RX_IDONE)) != (SPI_TX_IDONE|SPI_RX_IDONE));

	hw->flags &= ~(SPI_TX_IDONE|SPI_RX_IDONE);
	xfer = hw->current_transfer;
	msg = list_entry(hw->queue.next, struct spi_message, queue);

	len = hw->dma_xfer_len;
	if (len > BUFFER_SIZE)
		len = BUFFER_SIZE;

	/*
	 * If the rx buffer wasn't aligned, we used a bounce
	 * buffer for the transfer. Copy the data back and
	 * make the bounce buffer ready for re-use.
	 */
	if (xfer->rx_buf && xfer->rx_dma == INVALID_DMA_ADDRESS) {
		dma_sync_single_for_cpu(&hw->pdev->dev, hw->buf_rx_dma,
					len, DMA_FROM_DEVICE);
		memcpy(xfer->rx_buf, hw->buf_rx + xfer->len - hw->remaining_bytes, len);
	}

	hw->remaining_bytes -= len;

	if (hw->remaining_bytes == 0) {
		msg->actual_length += xfer->len;
 
		if (!msg->is_dma_mapped) {
			if (xfer->tx_dma != INVALID_DMA_ADDRESS)
				dma_unmap_single(master->cdev.dev,
						 xfer->tx_dma,
						 xfer->len,
						 DMA_TO_DEVICE);
			if (xfer->rx_dma != INVALID_DMA_ADDRESS)
				dma_unmap_single(master->cdev.dev,
						 xfer->rx_dma,
						 xfer->len,
						 DMA_FROM_DEVICE);
		}

		/* REVISIT: udelay in irq is unfriendly */
		if (xfer->delay_usecs)
			udelay(xfer->delay_usecs);

		if (msg->transfers.prev == &xfer->transfer_list) {

			/* Message complete */
			ns9xx0_spi_chipsel(msg->spi,
					   msg->spi->chip_select,
					   SPI_CS_INACTIVE);

			list_del(&msg->queue);
			msg->status = 0;

			dev_dbg(master->cdev.dev,
				"xfer complete: %u bytes transferred\n",
				msg->actual_length);

			spin_unlock(&hw->lock);
			msg->complete(msg->context);
			spin_lock(&hw->lock);

			hw->current_transfer = NULL;

			/* continue; complete() may have queued requests */
			if (list_empty(&hw->queue) || hw->stopping) {
				ns9xxx_disable_dma_ch(0, hw->dma_rx);
				ns9xxx_disable_dma_ch(0, hw->dma_tx);
			}
			else
				ns9xxx_spi_next_message(master);
		} else {
			if (xfer->cs_change) {
				ns9xx0_spi_chipsel(msg->spi,
						   msg->spi->chip_select,
						   SPI_CS_INACTIVE);
				udelay(1);
				ns9xx0_spi_chipsel(msg->spi,
						   msg->spi->chip_select,
						   SPI_CS_ACTIVE);
			}

			ns9xxx_spi_next_xfer(master, msg);
		}
	} else {
		/* There are still pending data in the current transfer */
		ns9xxx_spi_next_xfer(master, msg);
	}
}

static irqreturn_t ns9xxx_spi_irq_tx(int irq, void *dev, struct pt_regs *regs)
{
	struct spi_master	*master = dev;
	struct ns9xxx_spi	*hw = spi_master_get_devdata(master);
	u32			status;

	spin_lock(&hw->lock);

	/* Get status and mask all pending interrupts */
	status = __raw_readl( addr_dma_stat(0, hw->dma_tx));
	__raw_writel(status, addr_dma_stat(0, hw->dma_tx));

	dev_dbg(master->cdev.dev, "%s irq, status 0x%08x\n", __FUNCTION__, status);

	if (status & (NS_BBUS_DMA_STAT_ECIP|NS_BBUS_DMA_STAT_CAIP|NS_BBUS_DMA_STAT_PCIP)) {
		if (status & NS_BBUS_DMA_STAT_ECIP) {
			dev_err(master->cdev.dev, 
				"tx dma, Error Completion Irq, status 0x%08x\n", status);
		}
		if (status & NS_BBUS_DMA_STAT_CAIP) {
			dev_err(master->cdev.dev,
				"tx dma, Channel Abort Irq, status 0x%08x\n", status);
		}
		if (status & NS_BBUS_DMA_STAT_PCIP) {
			dev_err(master->cdev.dev, 
				"tx dma, Premature Complete Irq, status 0x%08x\n", status);
		}
		/* Retrigger the xfer */
		ns9xxx_spi_next_xfer(master, list_entry(hw->queue.next,
				     struct spi_message, queue));
	}

	if (status & NS_BBUS_DMA_STAT_NCIP) {
		hw->flags |= SPI_TX_IDONE;
		if (hw->flags & SPI_RX_IDONE)
			ns9xxx_spi_irq_trigger_next(master);
	}

	spin_unlock(&hw->lock);

	return IRQ_HANDLED;
}

static irqreturn_t ns9xxx_spi_irq_rx(int irq, void *dev, struct pt_regs *regs)
{
	struct spi_master	*master = dev;
	struct ns9xxx_spi	*hw = spi_master_get_devdata(master);
	u32			status, dlen;

	spin_lock(&hw->lock);

	/* Get status and mask all pending interrupts */
	status = __raw_readl(addr_dma_stat(0, hw->dma_rx));
	__raw_writel(status, addr_dma_stat(0, hw->dma_rx));

	dev_dbg(master->cdev.dev, "%s irq, status 0x%08x\n", __FUNCTION__, status);

	dlen = status & NS_BBUS_DMA_STAT_BLEN_MA;
	if (hw->dma_xfer_len != dlen) {
		dev_dbg(master->cdev.dev, "Incomplete dma xfer %d/%d\n",
			hw->dma_xfer_len, dlen);
		/* ?? what to do... ?? */
	}

	if (status & (NS_BBUS_DMA_STAT_ECIP|NS_BBUS_DMA_STAT_CAIP|NS_BBUS_DMA_STAT_PCIP)) {
		if (status & NS_BBUS_DMA_STAT_ECIP) {
			dev_err(master->cdev.dev, 
				"rx dma, Error Completion Irq, status 0x%08x\n", status);
		}
		if (status & NS_BBUS_DMA_STAT_CAIP) {
			dev_err(master->cdev.dev,
				"rx dma, Channel Abort Irq, status 0x%08x\n", status);
		}
		if (status & NS_BBUS_DMA_STAT_PCIP) {
			dev_err(master->cdev.dev, 
				"rx dma, Premature Complete Irq, status 0x%08x\n", status);
		}
		/* Retrigger the xfer */
		ns9xxx_spi_next_xfer(master, list_entry(hw->queue.next,
				     struct spi_message, queue));
	}

	if (status & NS_BBUS_DMA_STAT_NCIP) {
		hw->flags |= SPI_RX_IDONE;
		if (hw->flags & SPI_TX_IDONE)
			ns9xxx_spi_irq_trigger_next(master);
	}

	spin_unlock(&hw->lock);

	return IRQ_HANDLED;
}

static int __devinit ns9xxx_spi_probe(struct platform_device *pdev)
{
	struct resource		*regs;
	int			irq;
	int			ret = -ENOMEM;
	struct spi_master	*master;
	struct ns9xxx_spi	*hw;
	struct ns9xxx_spi_cfg	*pdata;


	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs) {
		dev_err(&pdev->dev, "Cannot get IORESOURCE_MEM\n");
		return -ENOENT;
	}	

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "Cannot get IRQ\n");
		return -ENOENT;
	}

	pdata = pdev->dev.platform_data;
	if (!pdata) {
		dev_err(&pdev->dev, "No platform data supplied\n");
		return -ENOENT;
	}

	/* setup spi core and ns9xxx-specific */
	master = spi_alloc_master(&pdev->dev, sizeof *hw);
	if (!master) {
		dev_err(&pdev->dev, "No memory for spi_master\n");
		goto out_ret;
	}
	master->bus_num = pdev->id;
	master->num_chipselect = 1;
	master->setup = ns9xxx_spi_setup;
	master->transfer = ns9xxx_spi_transfer;
	master->cleanup = ns9xxx_spi_cleanup;
	platform_set_drvdata(pdev, master);

	hw = spi_master_get_devdata(master);
	hw->master = master;
	hw->pdata = pdata;
	hw->flags = 0;
	hw->irq_rx = irq;
	hw->irq_tx = irq + 1;
	hw->dma_rx = master->bus_num * 2 + 1;
	hw->dma_tx = hw->dma_rx + 1;
	hw->pdev = pdev;
	spin_lock_init(&hw->lock);
	INIT_LIST_HEAD(&hw->queue);

	if (hw->pdata->use_dma == 0) {
		dev_err(&pdev->dev, "Standard IRQ, not supported yet, use DMA\n");
		ret = -ENODEV;
		goto out_free;
	} else {
		if ((ret = ns9xxx_spi_alloc_dma(hw)) != 0)
			goto out_free;
	}

	hw->ioarea = request_mem_region(regs->start, (regs->end - regs->start) + 1, pdev->name);
	if (!hw->ioarea) {
		dev_err(&pdev->dev, "Cannot reserve region\n");
		ret = -ENXIO;
		goto out_free_dma;
	}

	hw->regs = ioremap(regs->start, (regs->end - regs->start) + 1);
	if (!hw->regs) {
		dev_err(&pdev->dev, "Cannot map IO\n");
		ret = -ENXIO;
		goto out_no_iomap;
	}

	ret = request_irq(hw->irq_rx, ns9xxx_spi_irq_rx, 0, pdev->name, master);
	if (ret) {
		dev_err(&pdev->dev, "Cannot request IRQ %d\n", hw->irq_rx);
		goto out_unmap_regs;
	}

	ret = request_irq(hw->irq_tx, ns9xxx_spi_irq_tx, 0, pdev->name, master);
	if (ret) {
		dev_err(&pdev->dev, "Cannot request IRQ %d\n", hw->irq_tx);
		goto out_free_irq_rx;
	}

	/* Initialize the hardware */
	ret = ns9xx0_spi_request_gpios(hw);
	if (ret) {
		dev_err(&pdev->dev, "Cannot request GPIOs\n");
		goto out_free_irq_tx;
	}

	dev_info(&pdev->dev, "NS9XXX SPI Controller at 0x%08lx (irq_rx %d, irq_tx %d)" \
		 " (bbus dma_rx %d, dma_tx %d)\n", (unsigned long)regs->start, hw->irq_rx,
		 hw->irq_tx, hw->dma_rx, hw->dma_tx);

	ret = spi_register_master(master);
	if (ret) {
		dev_err(&pdev->dev, "Cannot register SPI master\n");
		goto out_reset_hw;
	}

	return 0;

out_reset_hw:
	cpu_writel(hw, SER_CTRLA, 0x0);
	ns9xx0_spi_release_gpios(hw);
out_free_irq_tx:
	free_irq(hw->irq_tx, master);
out_free_irq_rx:
	free_irq(hw->irq_rx, master);
out_unmap_regs:
	iounmap(hw->regs);
out_no_iomap:
	release_resource(hw->ioarea);
	kfree(hw->ioarea);
out_free_dma:
	ns9xxx_spi_free_dma(hw);
out_free:
	spi_master_put(master);
out_ret:
	return ret;
}


static int ns9xxx_spi_remove(struct platform_device *pdev)
{
	struct spi_master	*master = platform_get_drvdata(pdev);
	struct ns9xxx_spi 	*hw = spi_master_get_devdata(master);
	struct spi_message	*msg;

	/* reset the hardware and block queue progress */
	spin_lock_irq(&hw->lock);
	hw->stopping = 1;
	cpu_writel(hw, SER_CTRLA, 0x0);
	spin_unlock_irq(&hw->lock);

	/* Terminate remaining queued transfers */
	list_for_each_entry(msg, &hw->queue, queue) {
		msg->status = -ESHUTDOWN;
		msg->complete(msg->context);
	}

	ns9xxx_spi_free_dma(hw);

	free_irq(hw->irq_rx, master);
	free_irq(hw->irq_tx, master);
	iounmap(hw->regs);

	release_resource(hw->ioarea);
	kfree(hw->ioarea);

	ns9xx0_spi_release_gpios(hw);

	spi_unregister_master(master);
	spi_master_put(hw->master);

	return 0;
}


#ifdef CONFIG_PM

static int ns9xxx_spi_suspend(struct platform_device *pdev, pm_message_t msg)
{
	return 0;
}

static int ns9xxx_spi_resume(struct platform_device *pdev)
{
	return 0;
}

#else
#define ns9xxx_spi_suspend NULL
#define ns9xxx_spi_resume  NULL
#endif

static struct platform_driver ns9xxx_spidrv = {
	.probe		= ns9xxx_spi_probe,
	.remove		=  __devexit_p(ns9xxx_spi_remove),
	.suspend	= ns9xxx_spi_suspend,
	.resume		= ns9xxx_spi_resume,
	.driver		= {
		.name	= "ns9xxx_spi",
		.bus	= &platform_bus_type,
		.owner	= THIS_MODULE,
	},
};

static int __init ns9xxx_spi_init(void)
{
	return platform_driver_register(&ns9xxx_spidrv);
}

static void __exit ns9xxx_spi_exit(void)
{
        platform_driver_unregister(&ns9xxx_spidrv);
}

module_init(ns9xxx_spi_init);
module_exit(ns9xxx_spi_exit);

MODULE_DESCRIPTION("NS9XXX SPI Driver");
MODULE_AUTHOR("Digi International Inc.");
MODULE_LICENSE("GPL");
