/*
 *  linux/drivers/spi/spi-alloc.c
 *
 *  Copyright (C) 2005 MontaVista Software, Inc <sources@mvista.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/kmod.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/spi.h>
#include <asm/atomic.h>
#include "spi-core.h"

#define SPIMSG_POOL_SIZE	0x10000

static struct spi_msg_pool {
	void *vaddr;
	void *cur;
	int count;
} spimsg_pool;

static spinlock_t spimsg_lock = SPIN_LOCK_UNLOCKED;

struct spi_msg *__spimsg_kzalloc(void)
{
	struct spi_msg *msg = NULL;
	unsigned long flags;

	spin_lock_irqsave(&spimsg_lock, flags);
	if (spimsg_pool.count > 4) {
		msg = *(struct spi_msg **) spimsg_pool.cur;
		*(void **)spimsg_pool.cur = **(void ***)spimsg_pool.cur;
		memset(msg, 0, sizeof(*msg));
		spimsg_pool.count--;
	}
	spin_unlock_irqrestore(&spimsg_lock, flags);

	return msg;
}

void __spimsg_kfree(struct spi_msg *msg)
{
	unsigned long flags;

	if (msg) {
		if ((unsigned long)((long)msg - (long)spimsg_pool.vaddr) >
				SPIMSG_POOL_SIZE) {
			printk(KERN_ERR "Trying to free entry not from the SPI pool\n");
			BUG();
		}

		spin_lock_irqsave(&spimsg_lock, flags);
		*(long *)msg = *(long *)spimsg_pool.cur;
		*(long *)spimsg_pool.cur = (long)msg;
		spimsg_pool.count++;
		spin_unlock_irqrestore(&spimsg_lock, flags);
	}
}

int __spimsg_init(void)
{
	int ret = 0, i;

	int size = ((sizeof(struct spi_msg) + 3) >> 2) << 2;

	spimsg_pool.cur = spimsg_pool.vaddr =
		kzalloc(SPIMSG_POOL_SIZE, SLAB_KERNEL);

	if (!spimsg_pool.cur) {
		printk(KERN_ERR "Couldn't allocate large buffer for SPI\n");
		ret = -ENOMEM;
		goto out;
	}

	spimsg_pool.count = SPIMSG_POOL_SIZE / size;
	for (i = 0; i < spimsg_pool.count - 1; i++) {
		void **addr = spimsg_pool.vaddr + i*size;
		*addr = (void *)addr + size;
	}
	*(long *)(spimsg_pool.vaddr + (spimsg_pool.count - 1) * size) =
	    (long)spimsg_pool.vaddr;

out:
	return ret;
}

void __spimsg_exit(void)
{
	if (spimsg_pool.vaddr)
		kfree(spimsg_pool.vaddr);
}
