/*
 *  drivers/spi/spi-core.c
 *
 *  Copyright (C) 2005 MontaVista Software, Inc <sources@mvista.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
#include <linux/kmod.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/kthread.h>
#include <linux/spi.h>
#include <asm/atomic.h>

#include "spi-thread.h"
#include "spi-alloc.h"
#include "spi-core.h"

static int spi_device_del(struct device *dev, void *data);

/**
 * spimsg_set_clock - set message's clock
 * @message: SPI message
 * @clock: clock settings
 */
void spimsg_set_clock (struct spi_msg* message, u32 clock)
{
	message->clock = clock;
}

/**
 * spimsg_get_clock - get message's clock
 * @message: SPI message
 */
u32 spimsg_get_clock (struct spi_msg* message)
{
	return message->clock;
}

/**
 * spimsg_get_flags - get message's flags
 * @message: SPI message
 */
u32 spimsg_get_flags (struct spi_msg* message)
{
	return message->flags;
}

/**
 * spimsg_get_buffer - get the buffer for I/O ops
 * @message: SPI message
 * @buffer: ptr to the buffer to be filled in
 */
u32 spimsg_get_buffer (struct spi_msg *message, void **buffer)
{
	struct spi_bus_driver *bdrv = SPI_DEV_DRV(message->device);
	
	if (!buffer)
		return 0;

	*buffer = message->buf_ptr;

	if (message->flags & SPI_M_DMAUNSAFE && bdrv->dma_safe_alloc) {
		*buffer = bdrv->dma_safe_alloc (message->buf_ptr, message->len, message->flags & (SPI_M_RD|SPI_M_WR));
		if (!*buffer)
			return 0;
	}
	return message->len;
}

/**
 * spimsg_put_buffer - put the buffer used by I/O ops
 * @message: SPI message
 * @buffer: buffer used for I/O
 */
void spimsg_put_buffer (struct spi_msg *message, void *buffer)
{
	struct spi_bus_driver *bdrv = SPI_DEV_DRV(message->device);

	if (message->flags & SPI_M_DMAUNSAFE && bdrv->dma_safe_free)
		bdrv->dma_safe_free (message->buf_ptr, buffer, message->len, message->flags & (SPI_M_RD|SPI_M_WR));
}

/**
 * spimsg_get_spidev - obtain the SPI device from the message
 * @message: SPI message
 */
struct spi_device *spimsg_get_spidev (struct spi_msg* message)
{
	return message->device;
}

void spimsg_set_ctx(struct spi_msg *message, void *ctx)
{
	message->context = ctx;
}

void *spimsg_get_ctx(struct spi_msg *message)
{
	return message->context;
}

int spimsg_complete(struct spi_msg *message, int code)
{
	if (message->status)
		message->status(message, code);
	return code;
}

static inline struct spi_msg *__spimsg_alloc(struct spi_device *device,
					   struct spi_msg *link,
					   u32 flags,
					   void *buf,
					   unsigned short len,
					   void (*status) (struct spi_msg *,
							   int code))
{
	struct spi_msg *msg;

	if ((flags & (SPI_M_RD|SPI_M_WR)) == (SPI_M_RD|SPI_M_WR))
		return NULL;
	msg = spimsg_kzalloc();
	if (!msg)
		return NULL;
	msg->len = len;
	msg->status = status;
	msg->device = device;
	msg->flags = flags;
	INIT_LIST_HEAD(&msg->link);

	msg->buf_ptr = buf;

	if (link)
		link->next = msg;

	return msg;
}

/**
 * spimsg_alloc - allocate the spi message
 * @device: target device
 * @flags: SPI message flags
 * @buf: user-supplied buffer
 * @len: buffer's length
 * @status: user-supplied callback function
 */
struct spi_msg *spimsg_alloc(struct spi_device *device,
					   u32 flags,
					   void *buf,
					   unsigned short len,
					   void (*status) (struct spi_msg *,
							   int code))
{
	return __spimsg_alloc(device, NULL, flags, buf, len, status);
}

/**
 * spimsg_chain - allocate the spi message and link it to the current
 * @msg: current message
 * @flags: SPI message flags
 * @buf: user-supplied buffer
 * @len: buffer's length
 * @status: user-supplied callback function
 */
struct spi_msg *spimsg_chain(struct spi_msg *msg,
					   u32 flags,
					   void *buf,
					   unsigned short len,
					   void (*status) (struct spi_msg *,
							   int code))
{
	return msg ?
		__spimsg_alloc(msg->device, msg, flags, buf, len, status) :
		NULL;
}

/**
 * spimsg_free - free the message(s) allocated by spimsg_alloc
 * @msg: head message to free
 */
void spimsg_free(struct spi_msg *msg)
{
	if (msg->next)
		spimsg_free(msg->next);
	spimsg_kfree(msg);
}

/**
 * spimsg_getnext - get next message in chain
 * @message: this message
 */
struct spi_msg *spimsg_getnext(struct spi_msg *message)
{
	return message ? message->next : ERR_PTR(-EINVAL);
}

/**
 * spi_bus_match_name - verify that driver matches device on spi bus
 * @dev: device that hasn't yet being assigned to any driver
 * @drv: driver for spi device
 * Description:
 * 	match the device to driver.Drivers and devices on SPI bus
 * 	are matched by name, just like the platform devices
 */
static int spi_bus_match_name(struct device *dev, struct device_driver *drv)
{
	return !strcmp(TO_SPI_DEV(dev)->name, drv->name);
}

/**
 * spi_bus_suspend - suspend all devices on the spi bus
 * @dev: spi device to be suspended
 * @message: PM message
 * Description:
 * 	this function set device on SPI bus to suspended state, just
 * 	like platform_bus does
 */
static int spi_bus_suspend(struct device * dev, pm_message_t message)
{
	int ret = 0;

	if (dev->driver && TO_SPI_DRIVER(dev->driver)->suspend ) {
		ret = TO_SPI_DRIVER(dev->driver)->suspend( TO_SPI_DEV(dev), message);
		if (ret == 0 )
			dev->power.power_state = message;
	}
	return ret;
}

/**
 * spi_bus_resume - resume functioning of all devices on spi bus
 * @dev: device to resume
 * Description:
 * 	This function resumes device on SPI bus, just like platform_bus does
 */
static int spi_bus_resume(struct device * dev)
{
	int ret = 0;

	if (dev->driver && TO_SPI_DRIVER(dev->driver)->suspend ) {
		ret = TO_SPI_DRIVER(dev->driver)->resume(TO_SPI_DEV(dev));
		if (ret == 0)
			dev->power.power_state = PMSG_ON;
	}
	return ret;
}

struct bus_type spi_bus = {
	.name = "spi",
	.match = spi_bus_match_name,
	.suspend = spi_bus_suspend,
	.resume = spi_bus_resume,
};

/**
 * spi_bus_driver_init - init internal bus driver structures
 * @bus: registered spi_bus_driver structure
 * @dev: device that represents spi controller
 * Description:
 * 	Once registered by spi_bus_register, the bus driver needs
 * 	initialization, that includes starting thread, initializing
 * 	internal structures.. The best place where the spi_bus_driver_init
 * 	is in the `probe' function, when we already sure that passed
 * 	device object is SPI master controller.
 */
int spi_bus_driver_init(struct spi_bus_driver *bus, struct device *dev)
{
	struct spi_bus_data *pd =
	    kmalloc(sizeof(struct spi_bus_data), SLAB_KERNEL);
	int err = 0;

	if (!pd) {
		err = -ENOMEM;
		goto init_failed;
	}

	pd->bus = bus;
	pd->lock = SPIN_LOCK_UNLOCKED;
	INIT_LIST_HEAD(&pd->msgs);
	pd->id = dev->bus_id;

	if (!bus->start_async && !bus->stop_async) {
		bus->start_async = spi_start_async;
		bus->stop_async = spi_stop_async;
		if (!bus->queue)
			bus->queue = spi_queue;
	}

	dev->platform_data = pd;

	pd->async_data = bus->start_async ? bus->start_async(dev) : NULL;

	return 0;

init_failed:
	return err;
}

/**
 * spi_bus_driver_cleanup -  cleanup internal driver structures
 *
 * @bus: pointer to spi_bus_driver structure to deinitialize
 * @dev: bus device object
 *
 * Description: in order to roll back initialization of bus driver,
 * which is made by spi_bus_driver_init, this function should be used
 **/
int spi_bus_driver_cleanup (struct spi_bus_driver *bus_drv, struct device *dev)
{
	struct spi_bus_data *pd = dev->platform_data;

	if (!pd) {
		if(bus_drv->stop_async)
			bus_drv->stop_async(dev, pd->async_data);
		kfree (pd);
		dev->platform_data = NULL;
		device_for_each_child(dev, NULL, spi_device_del);
	}
	return 0;
}
/**
 * __spi_bus_free -- unregister all children of the spi bus
 * @dev: the spi bus `device' object
 * @context: not used, will be NULL
 * Description:
 * 	This is an internal function that is called when unregistering
 * 	bus driver. Responsibility of this function is freeing the
 * 	resources that were requested by spi_bus_driver_init
 */
static int __spi_bus_free(struct device *dev, void *context)
{
	return spi_bus_driver_cleanup(TO_SPI_BUS_DRIVER(dev->driver), dev);
}

/**
 * spi_bus_driver_unregister - unregister SPI bus controller from the system
 * @bus_driver: driver registered by call to spi_bus_driver_register
 * Description:
 * 	This routine unregisters the SPI bus from the system. Before
 * 	unregistering, it deletes each SPI device on the bus using call
 * 	to __spi_device_free
 */
void spi_bus_driver_unregister(struct spi_bus_driver *bus_driver)
{
	if (bus_driver) {
		driver_for_each_device(&bus_driver->driver, NULL, NULL, __spi_bus_free);
		driver_unregister(&bus_driver->driver);
	}
}

/**
 * spi_device_release - release the spi device structure
 * @dev: spi_device to be released
 * Description:
 * 	Pointer to this function will be put to dev->release place
 * 	This fus called as a part of device removing
 */
void spi_device_release(struct device *dev)
{
	struct spi_device* sdev = TO_SPI_DEV(dev);

	kfree(sdev);
}

/**
 * spi_device_add - add the new (discovered) SPI device to the bus. Mostly used by bus drivers
 * @parent: the bus device object
 * @name: name of device (non-null!)
 * @bus_data: bus data to be assigned to device
 * Description:
 * 	SPI device usually cannot be discovered by SPI bus driver, so it
 * 	needs to take the configuration somewhere from hardcoded structures,
 * 	and prepare bus_data for its devices
 */
struct spi_device* spi_device_add(struct device *parent, char *name, void *bus_data)
{
	struct spi_device* dev;
	static int minor = 0;

	if (!name)
		goto dev_add_out;

	dev = kzalloc(sizeof(struct spi_device), SLAB_KERNEL);
	if(!dev)
		goto dev_add_out;

	dev->dev.parent = parent;
	dev->dev.bus = &spi_bus;
	strncpy(dev->name, name, sizeof(dev->name));
	strncpy(dev->dev.bus_id, name, sizeof(dev->dev.bus_id));
	dev->dev.release = spi_device_release;
	dev->dev.platform_data = bus_data;

	if (device_register(&dev->dev)<0) {
		dev_dbg(parent, "device '%s' cannot be added\n", name);
		goto dev_add_out_2;
	}
	dev->cdev = spi_class_device_create(minor, &dev->dev);
	dev->minor = minor++;
	return dev;

dev_add_out_2:
	kfree(dev);
dev_add_out:
	return NULL;
}

/**
 * spi_device_del - delete the SPI device
 * @dev: device to delete
 * @data: data associated with the device
 */
static int spi_device_del(struct device *dev, void *data)
{
	struct spi_device *spidev = TO_SPI_DEV(dev);
	if (spidev->cdev) {
		spi_class_device_destroy(spidev->cdev);
		spidev->cdev = NULL;
	}
	device_unregister(&spidev->dev);
	return 0;
}
/**
 * __spi_transfer_callback - callback to process synchronous messages
 * @msg: message that is about to complete
 * @code: message status
 * Description:
 * 	callback for synchronously processed message. If spi_transfer
 * 	determines that there is no callback provided neither by
 * 	msg->status nor callback parameter, the __spi_transfer_callback
 * 	will be used, and spi_transfer does not return until transfer
 * 	is finished
 */
static void __spi_transfer_callback(struct spi_msg *msg, int code)
{
	complete(&msg->sync);
}

/*
 * spi_transfer - transfer the message either in sync or async way
 * @msg: message to process
 * @callback: user-supplied callback
 * @return:
 * 	if both msg->status and callback are set, the error code of -EINVAL
 * 	will be returned
 */
int spi_transfer(struct spi_msg *msg, void (*callback) (struct spi_msg *, int))
{
	int err = -EINVAL;
	struct device *bus = msg->device->dev.parent;

	if (msg && TO_SPI_BUS_DRIVER(bus->driver)->queue)
	{
		if (callback && !msg->status) {
			msg->status = callback;
			callback = NULL;
		}

		if (!callback) {
			if (!msg->status) {
				init_completion (&msg->sync);
				msg->status = __spi_transfer_callback;
				err = TO_SPI_BUS_DRIVER(bus->driver)->queue(msg);
				wait_for_completion(&msg->sync);
				err = 0;
			} else {
				err = TO_SPI_BUS_DRIVER(bus->driver)->queue(msg);
			}
		}
	}

	return err;
}

/**
 * spi_write - send data to a device on an SPI bus
 * @dev: the target device
 * @flags: additional flags for message
 * @buf: buffer to be sent
 * @len: buffer's length
 * @return: Returns the number of bytes transferred, or negative error code.
 */
int spi_write(struct spi_device *dev, u32 flags, char *buf, size_t len)
{
	struct spi_msg *msg = spimsg_alloc(dev, SPI_M_WR | SPI_M_DMAUNSAFE | flags, buf, len, NULL);
	int ret;

	ret = spi_transfer(msg, NULL);
	return ret == 1 ? len : ret;
}

/**
 * spi_read - receive data from a device on an SPI bus
 * @dev: the target device
 * @flags: additional flags for message
 * @buf: buffer to be sent
 * @len: buffer's length
 * @return: Returns the number of bytes transferred, or negative error code.
 */
int spi_read(struct spi_device *dev, u32 flags, char *buf, size_t len)
{
	int ret;
	struct spi_msg *msg = spimsg_alloc(dev, SPI_M_RD | SPI_M_DMAUNSAFE | flags, buf, len, NULL);

	ret = spi_transfer(msg, NULL);
	return ret == 1 ? len : ret;
}

/**
 * spi_bus_populate - populate the bus
 * @parent: the SPI bus device object
 * @devices_s: array of structures that represents bus population
 * @callback: optional pointer to function that called on each device's add
 * Description:
 * 	This function is intended to populate the SPI bus corresponding to
 * 	device passed as 1st parameter. If some device cannot be added
 * 	because of spi_device_add fail, the function will not try to parse
 * 	the rest of list
 */
int spi_bus_populate(struct device *parent,
			struct spi_device_desc* devices_s,
			void (*callback) (struct device* bus,
					  struct spi_device *new_dev,
					  void* params))
{
	struct spi_device *new_device;
	int count = 0;

	while (devices_s->name) {
		dev_dbg(parent, " discovered new SPI device, name '%s'\n",
				devices_s->name);
		if ((new_device = spi_device_add(parent, devices_s->name, devices_s->params)) == NULL)
			break;
		if (callback)
			callback(parent, new_device, devices_s->params);
		devices_s++;
		count++;
	}
	return count;
}

/**
 * spi_bus_reset - reset the spi bus
 * @bus: device object that represents the SPI bus
 * @context: u32 value to be passed to reset method of bus
 * Description:
 * 	This is simple wrapper for bus' `reset' method
 */
void spi_bus_reset (struct device *bus, u32 context)
{
	if (bus && bus->driver && TO_SPI_BUS_DRIVER(bus->driver)->reset)
		TO_SPI_BUS_DRIVER(bus->driver)->reset(bus, context);
}

/*
 * The functions below are wrappers for corresponding device_driver's methods
 */
static int spi_driver_probe (struct device *dev)
{
	struct spi_driver *sdrv = TO_SPI_DRIVER(dev->driver);
	struct spi_device *sdev = TO_SPI_DEV(dev);

	return sdrv->probe ? sdrv->probe(sdev) : -EFAULT;
}

static int spi_driver_remove (struct device *dev)
{
	struct spi_driver *sdrv = TO_SPI_DRIVER(dev->driver);
	struct spi_device *sdev = TO_SPI_DEV(dev);

	return  sdrv->remove  ? sdrv->remove(sdev) : -EFAULT;
}

static void spi_driver_shutdown (struct device *dev)
{
	struct spi_driver *sdrv = TO_SPI_DRIVER(dev->driver);
	struct spi_device *sdev = TO_SPI_DEV(dev);

	if (dev->driver && sdrv->shutdown)
		sdrv->shutdown(sdev);
}

static int __init spi_core_init(void)
{
	int ret = spidev_init();

	if (ret < 0)
		goto out;

	ret = spimsg_init();
	if (ret < 0)
		goto out;

	ret = bus_register(&spi_bus);

out:
	return ret;
}

int spi_driver_add(struct spi_driver *drv)
{
	drv->driver.bus = &spi_bus;
	drv->driver.probe = spi_driver_probe;
	drv->driver.remove = spi_driver_remove;
	drv->driver.shutdown = spi_driver_shutdown;
	return driver_register(&drv->driver);
}

static void __exit spi_core_exit(void)
{
	bus_unregister(&spi_bus);
	spimsg_exit();
	spidev_cleanup();
}

subsys_initcall(spi_core_init);
module_exit(spi_core_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("dmitry pervushin <dpervushin@ru.mvista.com>");

EXPORT_SYMBOL_GPL(spi_bus_reset);
EXPORT_SYMBOL_GPL(spi_device_add);
EXPORT_SYMBOL_GPL(spi_driver_add);
EXPORT_SYMBOL_GPL(spi_bus_driver_unregister);
EXPORT_SYMBOL_GPL(spi_bus_populate);
EXPORT_SYMBOL_GPL(spi_transfer);
EXPORT_SYMBOL_GPL(spi_write);
EXPORT_SYMBOL_GPL(spi_read);
EXPORT_SYMBOL_GPL(spi_bus);
EXPORT_SYMBOL_GPL(spi_bus_driver_init);
EXPORT_SYMBOL_GPL(spi_bus_driver_cleanup);

EXPORT_SYMBOL_GPL(spimsg_alloc);
EXPORT_SYMBOL_GPL(spimsg_chain);
EXPORT_SYMBOL_GPL(spimsg_free);
EXPORT_SYMBOL_GPL(spimsg_put_buffer);
EXPORT_SYMBOL_GPL(spimsg_get_flags);
EXPORT_SYMBOL_GPL(spimsg_get_buffer);
EXPORT_SYMBOL_GPL(spimsg_get_clock);
EXPORT_SYMBOL_GPL(spimsg_set_clock);
EXPORT_SYMBOL_GPL(spimsg_getnext);
EXPORT_SYMBOL_GPL(spimsg_get_spidev);
EXPORT_SYMBOL_GPL(spimsg_set_ctx);
EXPORT_SYMBOL_GPL(spimsg_get_ctx);
EXPORT_SYMBOL_GPL(spimsg_complete);

