/*
 *  linux/include/linux/spi.h
 *
 *  Copyright (C) 2005 MontaVista Software, Inc <sources@mvista.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * Derived from l3.h by Jamey Hicks
 */
#ifndef SPI_H
#define SPI_H

#include <linux/types.h>
#include <linux/device.h>

struct spi_device;
struct spi_driver;
struct spi_msg;
struct spi_bus_driver;

extern struct bus_type spi_bus;

struct spi_bus_data {
	spinlock_t lock;
	struct list_head msgs;
	void *async_data;
	wait_queue_head_t queue;
	struct spi_device *selected_device;
	struct spi_bus_driver *bus;
	char *id;
};

#define TO_SPI_BUS_DRIVER(drv) container_of(drv, struct spi_bus_driver, driver)
struct spi_bus_driver {

	int 	(*xfer) (struct spi_msg * msg);
	void 	(*set_clock) (struct device * bus_device, u32 clock_hz);
	void 	(*reset) (struct device *bus_device, u32 context);

	int	(*queue) (struct spi_msg *msg);
	void   *(*start_async)( struct device *bus);
	void 	(*stop_async)( struct device *bus, void *async);
	void   *(*dma_safe_alloc)(void *buf, size_t len, int dir);
	void    (*dma_safe_free)(void *buf, void *safe, size_t len, int dir);

	struct device_driver driver;
};

#define TO_SPI_DEV(device) container_of(device, struct spi_device, dev)
struct spi_device {
	char name[BUS_ID_SIZE];
	int minor;
	struct class_device *cdev;
	struct device dev;
};

static inline void *spi_device_get_busdata(struct spi_device *dev)
{
	return dev->dev.platform_data;
}

#define TO_SPI_DRIVER(drv) container_of(drv, struct spi_driver, driver)
struct spi_driver {

	int     (*probe)        (struct spi_device * dev);
	int     (*remove)       (struct spi_device * dev);
	void    (*shutdown)     (struct spi_device * dev);
	int	(*suspend)	(struct spi_device * dev, pm_message_t pm);
	int 	(*resume)	(struct spi_device * dev);

	struct device_driver driver;
};

#define SPI_DEV_DRV(device)  TO_SPI_BUS_DRIVER((device)->dev.parent->driver)

#define spi_device_lock(spi_dev)	 down(&(spi_dev)->dev.sem)
#define spi_device_unlock(spi_dev)	 up(&(spi_dev)->dev.sem)

#define SPI_M_RD	0x00000001
#define SPI_M_WR	0x00000002	/**< Write mode flag */
#define SPI_M_CSREL	0x00000004	/**< CS release level at end of the frame  */
#define SPI_M_CS	0x00000008	/**< CS active level at begining of frame  */
#define SPI_M_CSKEEP	0x00000010	/**< Don't change CS */
#define SPI_M_CPOL	0x00000100	/**< Clock polarity */
#define SPI_M_CPHA	0x00000200	/**< Clock Phase */
#define SPI_M_DMAUNSAFE 0x10000000	/** buffer is dma-unsafe */

void spimsg_set_ctx (struct spi_msg *message, void *ctx);
void *spimsg_get_ctx (struct spi_msg *message);
int spimsg_complete (struct spi_msg *message, int code);
void spimsg_set_clock (struct spi_msg* message, u32 clock);
u32 spimsg_get_clock (struct spi_msg* message);
struct spi_device *spimsg_get_spidev (struct spi_msg* message);
u32 spimsg_get_flags (struct spi_msg* message);
u32 spimsg_get_buffer (struct spi_msg *message, void **buffer);
void spimsg_put_buffer (struct spi_msg *message, void *buffer);
struct spi_msg *spimsg_alloc(struct spi_device *device,
			   u32 flags,
			   void *buf,
			   unsigned short len,
			   void (*status) (struct spi_msg *,
					   int code));
struct spi_msg *spimsg_chain(struct spi_msg *msg,
			   u32 flags,
			   void *buf,
			   unsigned short len,
			   void (*status) (struct spi_msg *,
					   int code));

void spimsg_free (struct spi_msg *message);
struct spi_msg *spimsg_getnext(struct spi_msg *message);

#if defined (CONFIG_SPI_CHARDEV)
extern struct class_device *spi_class_device_create(int minor, struct device *device);
extern void spi_class_device_destroy(struct class_device *cdev);
#else
static inline struct class_device *spi_class_device_create(int minor, struct device *device)
{
	return NULL;
}
static inline void  spi_class_device_destroy(struct class_device *cdev)
{
}
#endif

enum {
	SPIMSG_OK = 0,
	SPIMSG_FAILED = -1,
};

#define SPI_MAJOR	153

struct spi_driver;
struct spi_device;

#if defined (CONFIG_SPI_CHARDEV)
extern int __init spidev_init(void);
extern void __exit spidev_cleanup(void);
#else
static inline int spidev_init(void)
{
	return 0;
}
static inline void spidev_cleanup(void)
{
}
#endif

static inline int spi_bus_driver_register (struct spi_bus_driver *bus_driver)
{
	return driver_register (&bus_driver->driver);
}

void spi_bus_driver_unregister(struct spi_bus_driver *);
int spi_bus_driver_init(struct spi_bus_driver *driver, struct device *dev);
int spi_bus_driver_cleanup(struct spi_bus_driver *driver, struct device *dev);
struct spi_device* spi_device_add(struct device *parent, char *name, void *private);

extern int spi_driver_add(struct spi_driver *drv);

static inline void spi_driver_del(struct spi_driver *drv)
{
	driver_unregister(&drv->driver);
}

extern void spi_bus_reset(struct device* bus, u32 context);
extern int spi_write(struct spi_device *dev, u32 flags, char *buf, size_t len);
extern int spi_read(struct spi_device *dev, u32 flags, char *buf, size_t len);

extern int spi_transfer(struct spi_msg *message,
			void (*status) (struct spi_msg *, int));
struct spi_device_desc {
	char* name;
	void* params;
};
extern int spi_bus_populate(struct device *parent,
			     struct spi_device_desc *devices,
			     void (*assign) (struct device *parent,
				             struct spi_device *,
					     void *));

static inline int spi_w8r8 (struct spi_device *dev, u8 wr)
{
	u8 byte;
	int rc = -ENOMEM;
	struct spi_msg *msg = spimsg_alloc(dev,
			SPI_M_WR | SPI_M_CS | SPI_M_DMAUNSAFE,
			&wr,
			1,
			NULL);
	if (!msg)
		goto out;

	if (!spimsg_chain(msg,
			SPI_M_RD | SPI_M_CSREL | SPI_M_DMAUNSAFE,
			&byte,
			1,
			NULL))
		goto out;

	rc = spi_transfer(msg, NULL);
	spimsg_free(msg);

out:
	return rc < 0 ? rc : byte;
}

static inline int spi_w8r16 (struct spi_device *dev, u8 wr)
{
	u16 word;
	int rc = -ENOMEM;
	struct spi_msg *msg = spimsg_alloc(dev,
			SPI_M_WR | SPI_M_CS | SPI_M_DMAUNSAFE,
			&wr,
			1,
			NULL);
	if (!msg)
		goto out;

	if (!spimsg_chain(msg,
			SPI_M_RD | SPI_M_CSREL | SPI_M_DMAUNSAFE,
			&word,
			2,
			NULL))
		goto out;

	rc = spi_transfer(msg, NULL);
	spimsg_free(msg);

out:
	return rc < 0 ? rc : word;
}

static inline int spi_sync(struct spi_msg *message)
{
	return spi_transfer(message, NULL);
}

static inline int spi_async(struct spi_msg *message, void (*status) (struct spi_msg *, int))
{
	return spi_transfer(message, status);
}

#endif				/* SPI_H */
